from vcx.common import do_call, create_cb
from ctypes import *
from vcx.api.vcx_base import VcxBase


class VcxStateful(VcxBase):

    def __init__(self, source_id: str):
        VcxBase.__init__(self, source_id)

    async def _update_state(self, cls, fn: str) -> int:
        if not hasattr(cls.update_state, "cb"):
            self.logger.debug("{}: Creating callback".format(fn))
            cls.update_state.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))

        c_handle = c_uint32(self.handle)

        state = await do_call(fn,
                              c_handle,
                              cls.update_state.cb)

        self.logger.debug("{} object has state of: {}".format(cls, state))
        return state

    async def _update_state_with_message(self, cls, message: str, fn: str) -> int:
        if not hasattr(cls.update_state_with_message, "cb"):
            self.logger.debug("{}: Creating callback".format(fn))
            cls.update_state_with_message.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))

        c_handle = c_uint32(self.handle)
        c_message = c_char_p(message.encode('utf-8'))

        state = await do_call(fn,
                              c_handle,
                              c_message,
                              cls.update_state_with_message.cb)

        self.logger.debug("{} object has state of: {}".format(cls, state))
        return state

    async def _get_state(self, cls, fn: str) -> int:
        if not hasattr(cls.get_state, "cb"):
            self.logger.debug("{}: Creating callback".format(fn))
            cls.get_state.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))


        c_handle = c_uint32(self.handle)

        return await do_call(fn,
                             c_handle,
                             cls.get_state.cb)



from ctypes import *
from vcx.common import do_call, create_cb
from vcx.api.connection import Connection
from vcx.api.vcx_stateful import VcxStateful

import json

class IssuerCredential(VcxStateful):
    """
    The object of the VCX API representing an Issuer side in the credential issuance process.
    Assumes that pairwise connection between Issuer and Holder is already established.

    # State

    The set of object states and transitions depends on communication method is used.
    The communication method can be specified as config option on one of *_init function. The default communication method us `proprietary`.

    proprietary:
        VcxStateType::VcxStateInitialized - once `vcx_issuer_create_credential` (create IssuerCredential object) is called.

        VcxStateType::VcxStateOfferSent - once `vcx_issuer_send_credential_offer` (send `CRED_OFFER` message) is called.

        VcxStateType::VcxStateRequestReceived - once `CRED_REQ` messages is received.
                                                use `vcx_issuer_credential_update_state` or `vcx_issuer_credential_update_state_with_message` functions for state updates.
        VcxStateType::VcxStateAccepted - once `vcx_issuer_send_credential` (send `CRED` message) is called.

    aries:
        VcxStateType::VcxStateInitialized - once `vcx_issuer_create_credential` (create IssuerCredential object) is called.

        VcxStateType::VcxStateOfferSent - once `vcx_issuer_send_credential_offer` (send `CredentialOffer` message) is called.

        VcxStateType::VcxStateRequestReceived - once `CredentialRequest` messages is received.
        VcxStateType::None - once `ProblemReport` messages is received.
                                                use `vcx_issuer_credential_update_state` or `vcx_issuer_credential_update_state_with_message` functions for state updates.

        VcxStateType::VcxStateAccepted - once `vcx_issuer_send_credential` (send `Credential` message) is called.

    # Transitions

    proprietary:
        VcxStateType::None - `vcx_issuer_create_credential` - VcxStateType::VcxStateInitialized

        VcxStateType::VcxStateInitialized - `vcx_issuer_send_credential_offer` - VcxStateType::VcxStateOfferSent

        VcxStateType::VcxStateOfferSent - received `CRED_REQ` - VcxStateType::VcxStateRequestReceived

        VcxStateType::VcxStateRequestReceived - `vcx_issuer_send_credential` - VcxStateType::VcxStateAccepted

    aries: RFC - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential
        VcxStateType::None - `vcx_issuer_create_credential` - VcxStateType::VcxStateInitialized

        VcxStateType::VcxStateInitialized - `vcx_issuer_send_credential_offer` - VcxStateType::VcxStateOfferSent

        VcxStateType::VcxStateOfferSent - received `CredentialRequest` - VcxStateType::VcxStateRequestReceived
        VcxStateType::VcxStateOfferSent - received `ProblemReport` - VcxStateType::None

        VcxStateType::VcxStateRequestReceived - vcx_issuer_send_credential` - VcxStateType::VcxStateAccepted

        VcxStateType::VcxStateAccepted - received `Ack` - VcxStateType::VcxStateAccepted

    # Messages

    proprietary:
        CredentialOffer (`CRED_OFFER`)
        CredentialRequest (`CRED_REQ`)
        Credential (`CRED`)

    aries:
        CredentialProposal - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#propose-credential
        CredentialOffer - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#offer-credential
        CredentialRequest - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#request-credential
        Credential - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#issue-credential
        ProblemReport - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0035-report-problem#the-problem-report-message-type
        Ack - https://github.com/hyperledger/aries-rfcs/tree/master/features/0015-acks#explicit-acks
    """

    def __init__(self, source_id: str, attrs: dict, cred_def_id: str, name: str, price: float):

        VcxStateful.__init__(self, source_id)
        self._cred_def_id = cred_def_id
        self._attrs = attrs
        self._name = name
        self._price = price

    def __del__(self):
        self.release()
        self.logger.debug("Deleted {} obj: {}".format(IssuerCredential, self.handle))

    @staticmethod
    async def create(source_id: str, attrs: dict, cred_def_handle: int, name: str, price: str):
        """
            Create a Issuer Credential object that provides a credential for an enterprise's user
            Assumes a credential definition has been already written to the ledger.

            :param source_id: Tag associated by user of sdk
            :param attrs: attributes that will form the credential
            :param cred_def_handle: Handle from previously created credential def object
            :param name: Name given to the Credential
            :param price: Price, in tokens, required as payment for the issuance of the credential.

            Example:
            source_id = '1'
            cred_def_handle = 1
            attrs = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}
            name = 'Credential Name'
            issuer_did = '8XFh8yBzrpJQmNyZzgoTqB'
            phone_number = '8019119191'
            price = 1
            issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_handle, name, price)
        """
        constructor_params = (source_id, attrs, cred_def_handle, name, price)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_cred_def_handle = c_uint32(cred_def_handle)
        c_price = c_char_p(price.encode('utf-8'))
        # default institution_did in config is used as issuer_did
        c_issuer_did = None
        c_data = c_char_p(json.dumps(attrs).encode('utf-8'))
        c_name = c_char_p(name.encode('utf-8'))
        c_params = (c_source_id, c_cred_def_handle, c_issuer_did, c_data, c_name, c_price)

        return await IssuerCredential._create("vcx_issuer_create_credential",
                                              constructor_params,
                                              c_params)

    @staticmethod
    async def deserialize(data: dict):
        """
            Create a IssuerCredential object from a previously serialized object
            :param data: dict representing a serialized IssuerCredential Object
            :return: IssuerCredential object

            Example:
            source_id = '1'
            cred_def_id = 'cred_def_id1'
            attrs = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}
            name = 'Credential Name'
            issuer_did = '8XFh8yBzrpJQmNyZzgoTqB'
            phone_number = '8019119191'
            price = 1
            issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
            data = await issuer_credential.serialize()
            issuer_credential2 = await IssuerCredential.deserialize(data)
        """
        issuer_credential = await IssuerCredential._deserialize("vcx_issuer_credential_deserialize",
                                                      json.dumps(data),
                                                      data.get('data').get('source_id'),
                                                      data.get('data').get('price'),
                                                      data.get('data').get('credential_attributes'),
                                                      data.get('data').get('schema_seq_no'),
                                                      data.get('data').get('credential_request'))
        return issuer_credential

    async def serialize(self) -> dict:
        """
            Serializes the  issuer credential object for storage and later deserialization.

            Example:
            source_id = '1'
            cred_def_id = 'cred_def_id1'
            attrs = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}
            name = 'Credential Name'
            issuer_did = '8XFh8yBzrpJQmNyZzgoTqB'
            phone_number = '8019119191'
            price = 1
            issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
            data = await issuer_credential.serialize()
        :return: dictionary representing the serialized object
        """
        return await self._serialize(IssuerCredential, 'vcx_issuer_credential_serialize')

    async def update_state(self) -> int:
        """
        Query the agency for the received messages.
        Checks for any messages changing state in the object and updates the state attribute.

        Example:
        issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
        issuer_credential.update_state()
        :return:
        """
        return await self._update_state(IssuerCredential, 'vcx_issuer_credential_update_state')

    async def update_state_with_message(self, message: str) -> int:
        """
        Update the state of the credential based on the given message.
        Example:
        cred = await IssuerCredential.create(source_id)
        assert await cred.update_state_with_message(message) == State.Accepted
        :param message: message to process for state changes
        :return Current state of the IssuerCredential
        """
        return await self._update_state_with_message(IssuerCredential, message, 'vcx_issuer_credential_update_state_with_message')

    async def get_state(self) -> int:
        """
        Get the current state of the issuer credential object
        Example:
        issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
        issuer_credential.update_state()
        :return: State of the Object. Possible states:
                                         1 - Initialized
                                         2 - Offer Sent
                                         3 - Request Received
                                         4 - Issued
        """
        return await self._get_state(IssuerCredential, 'vcx_issuer_credential_get_state')

    def release(self) -> None:
        """
        Used to release memory associated with this object, used by the c library.
        :return:
        """
        self._release(IssuerCredential, 'vcx_issuer_credential_release')

    async def send_offer(self, connection: Connection):
        """
        Send a credential offer to a holder showing what will be included in the actual credential
        :param connection: Connection that identifies pairwise connection
        :return: None

        Example:
        source_id = '1'
        cred_def_id = 'cred_def_id1'
        attrs = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}
        name = 'Credential Name'
        issuer_did = '8XFh8yBzrpJQmNyZzgoTqB'
        phone_number = '8019119191'
        price = 1
        issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
        connection = await Connection.create(source_id)
        issuer_credential.send_offer(connection)
        """
        if not hasattr(IssuerCredential.send_offer, "cb"):
            self.logger.debug("vcx_issuer_send_credential_offer: Creating callback")
            IssuerCredential.send_offer.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_credential_handle = c_uint32(self.handle)
        c_connection_handle = c_uint32(connection.handle)

        await do_call('vcx_issuer_send_credential_offer',
                      c_credential_handle,
                      c_connection_handle,
                      IssuerCredential.send_offer.cb)

    async def get_offer_msg(self):
        """
        Gets the offer message that can be sent to the specified connection
        :param connection: Connection that identifies pairwise connection
        :return: None

        Example:
        source_id = '1'
        cred_def_id = 'cred_def_id1'
        attrs = {'key': 'value', 'key2': 'value2', 'key3': 'value3'}
        name = 'Credential Name'
        issuer_did = '8XFh8yBzrpJQmNyZzgoTqB'
        phone_number = '8019119191'
        price = 1
        issuer_credential = await IssuerCredential.create(source_id, attrs, cred_def_id, name, price)
        connection = await Connection.create(source_id)
        issuer_credential.get_offer_msg(connection)
        """
        if not hasattr(IssuerCredential.get_offer_msg, "cb"):
            self.logger.debug("vcx_issuer_get_credential_offer_msg: Creating callback")
            IssuerCredential.get_offer_msg.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)

        msg = await do_call('vcx_issuer_get_credential_offer_msg',
                            c_credential_handle,
                            IssuerCredential.get_offer_msg.cb)

        return json.loads(msg.decode())

    async def send_credential(self, connection: Connection):
        """
        Sends the credential to the end user (holder).
        :param connection: Connection that identifies pairwise connection
        :return: None
            Example:
            credential.send_credential(connection)
        """
        if not hasattr(IssuerCredential.send_credential, "cb"):
            self.logger.debug("vcx_issuer_send_credential: Creating callback")
            IssuerCredential.send_credential.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_credential_handle = c_uint32(self.handle)
        c_connection_handle = c_uint32(connection.handle)

        await do_call('vcx_issuer_send_credential',
                      c_credential_handle,
                      c_connection_handle,
                      IssuerCredential.send_credential.cb)

    async def get_credential_msg(self, my_pw_did: str):
        """
        Get the credential to send to the end user (prover).
        :param my_pw_did: my pw did associated with person I'm sending credential to
        :return: None
            Example:
            credential.send_credential(connection)
        """
        if not hasattr(IssuerCredential.get_credential_msg, "cb"):
            self.logger.debug("vcx_issuer_get_credential_msg: Creating callback")
            IssuerCredential.get_credential_msg.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)
        c_my_pw_did = c_char_p(json.dumps(my_pw_did).encode('utf-8'))

        msg = await do_call('vcx_issuer_get_credential_msg',
                            c_credential_handle,
                            c_my_pw_did,
                            IssuerCredential.get_credential_msg.cb)

        return json.loads(msg.decode())

    async def revoke_credential(self):
        """
        Revokes a credential.
        :return: None
            Example:
            credential.revoke_credential()
        """
        if not hasattr(IssuerCredential.revoke_credential, "cb"):
            self.logger.debug("vcx_issuer_revoke_credential: Creating callback")
            IssuerCredential.revoke_credential.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_credential_handle = c_uint32(self.handle)

        await do_call('vcx_issuer_revoke_credential',
                      c_credential_handle,
                      IssuerCredential.revoke_credential.cb)

    async def get_payment_txn(self):
        """
        Retrieve the payment transaction associated with this credential. This can be used to get the txn that
        was used to pay the issuer from the prover.  This could be considered a receipt of payment from the payer to
        the issuer.

        Example:
        txn = credential.get_payment_txn()
        :return: payment transaction
          {
              "amount":25,
              "inputs":[
                  "pay:null:1_3FvPC7dzFbQKzfG",
                  "pay:null:1_lWVGKc07Pyc40m6"
              ],
              "outputs":[
                  {"recipient":"pay:null:FrSVC3IrirScyRh","amount":5,"extra":null},
                  {"recipient":"pov:null:OsdjtGKavZDBuG2xFw2QunVwwGs5IB3j","amount":25,"extra":null}
              ]
          }
        """
        if not hasattr(IssuerCredential.get_payment_txn, "cb"):
            self.logger.debug("vcx_issuer_credential_get_payment_txn: Creating callback")
            IssuerCredential.get_payment_txn.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)

        payment_txn = await do_call('vcx_issuer_credential_get_payment_txn',
                      c_credential_handle,
                      IssuerCredential.get_payment_txn.cb)

        return json.loads(payment_txn.decode())




from typing import Optional
from ctypes import *
from vcx.error import VcxError, ErrorCode
from vcx.api.vcx_base import VcxBase
from vcx.common import do_call, create_cb
from vcx.api.vcx_stateful import VcxStateful

import json


class CredentialDef(VcxStateful):
    """
    Object that represents a credential definition written on the ledger.

    Attributes:
        source_id: user generated unique identifier
        schema_id: the ledger ID of the schema
        name: name of the credential definition
        transaction: credential def transaction that must be published to the ledger by Endorser
    """


    def __init__(self, source_id: str, name: str, schema_id: str, transaction: Optional[str] = None):
        VcxStateful.__init__(self, source_id)
        self._source_id = source_id
        self._schema_id = schema_id
        self._name = name
        self._transaction = transaction


    def __del__(self):
        self.release()
        self.logger.debug("Deleted {} obj: {}".format(CredentialDef, self.handle))

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, x):
        self._name = x

    @property
    def schema_id(self):
        return self._schema_id

    @schema_id.setter
    def schema_id(self, x):
        self._schema_id = x


    @property
    def transaction(self):
        return self._transaction


    @transaction.setter
    def transaction(self, x):
        self._transaction = x


    @staticmethod
    async def create(source_id: str, name: str, schema_id: str, payment_handle: int):
        """
        Creates a new CredentialDef object that is written to the ledger

        :param source_id: Institution's unique ID for the credential definition
        :param name: Name of credential definition
        :param schema_id: The schema ID given during the creation of the schema
        :param payment_handle: NYI - payment of ledger fee is taken from wallet automatically
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        endorser = 'V4SGRU86Z58d6TV7PBUe6f'
        credential_def1 = await CredentialDef.prepare_for_endorser(source_id, name, schema_id, endorser)
        :return: credential_def object,
        credential_def transaction that should be passed to Endorser for witting to ledger
        """
        constructor_params = (source_id, name, schema_id)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_schema_id = c_char_p(schema_id.encode('utf-8'))
        c_name = c_char_p(name.encode('utf-8'))
        # default institution_did in config is used as issuer_did
        c_issuer_did = None
        c_payment = c_uint32(payment_handle)
        # Todo: add params for tag and config
        c_tag = c_char_p('tag1'.encode('utf-8'))
        c_config = c_char_p('{"support_revocation":false}'.encode('utf-8'))
        c_params = (c_source_id, c_name, c_schema_id, c_issuer_did, c_tag, c_config, c_payment)

        return await CredentialDef._create("vcx_credentialdef_create",
                                           constructor_params,
                                           c_params)


    @staticmethod
    async def prepare_for_endorser(source_id: str, name: str, schema_id: str, endorser: str):
        """
        Create a new CredentialDef object that will be published on the ledger by Endorser later.
        
        Note that CredentialDef can't be used for credential issuing until it will be published on the ledger.

        :param source_id: Institution's unique ID for the credential definition
        :param name: Name of credential definition
        :param schema_id: The schema ID given during the creation of the schema
        :param endorser: DID of the Endorser that will submit the transaction.
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        :return: credential_def object, written to ledger
        """
        try:
            credentialdef = CredentialDef(source_id, '', '')

            if not hasattr(CredentialDef.prepare_for_endorser, "cb"):
                credentialdef.logger.debug("vcx_prepare_for_endorser: Creating callback")
                CredentialDef.prepare_for_endorser.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32,
                                                                            c_char_p, c_char_p, c_char_p))

            c_source_id = c_char_p(source_id.encode('utf-8'))
            c_name = c_char_p(name.encode('utf-8'))
            c_schema_id = c_char_p(schema_id.encode('utf-8'))
            c_endorser = c_char_p(endorser.encode('utf-8'))
            c_issuer_did = None
            c_tag = c_char_p('tag1'.encode('utf-8'))
            c_config = c_char_p('{"support_revocation":false}'.encode('utf-8'))

            handle, transaction, _, _ = await do_call('vcx_credentialdef_prepare_for_endorser',
                                                      c_source_id,
                                                      c_name,
                                                      c_schema_id,
                                                      c_issuer_did,
                                                      c_tag,
                                                      c_config,
                                                      c_endorser,
                                                      CredentialDef.prepare_for_endorser.cb)
            credentialdef.logger.debug("created credential def object")

            credentialdef.name = name
            credentialdef.handle = handle
            credentialdef.schema_id = schema_id
            credentialdef.transaction = transaction
            return credentialdef
        except KeyError:
            raise VcxError(ErrorCode.InvalidCredentialDef)


    async def serialize(self) -> dict:
        """
        Serialize the object for storage

        :return: serialized object
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        data1 = await credential_def1.serialize()
        :return: A re-instantiated object
        """
        return await self._serialize(CredentialDef, 'vcx_credentialdef_serialize')

    @staticmethod
    async def deserialize(data: dict):
        """
        Create the object from a previously serialized object.

        :param data: The output of the "serialize" call
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        data1 = await credential_def1.serialize()
        credential_def2 = await CredentialDef.deserialize(data1)
        :return: A re-instantiated object
        """
        try:
            credential_def = await CredentialDef._deserialize("vcx_credentialdef_deserialize",
                                                              json.dumps(data),
                                                              data['data']['source_id'],
                                                              data['data']['name'],
                                                              data['data']['id'])
            return credential_def
        except KeyError:
            raise VcxError(ErrorCode.InvalidCredentialDef)


    async def get_cred_def_id(self):
        """
        Get the ledger ID of the object

        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        assert await credential_def.get_cred_def_id() == '2hoqvcwupRTUNkXn6ArYzs:3:CL:2471'
        :return: ID string
        """
        cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))
        c_handle = c_uint32(self.handle)
        cred_def_id = await do_call('vcx_credentialdef_get_cred_def_id', c_handle, cb)
        return cred_def_id.decode()

    def release(self) -> None:
        """
        destroy the object and release any memory associated with it
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        credential_def1.release()
        :return: None
        """
        self._release(CredentialDef, 'vcx_credentialdef_release')

    async def get_payment_txn(self):
        """
        Get the payment transaction information generated when paying the ledger fee

        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def1 = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        txn = await credential_def1.get_payment_txn()
        :return: JSON object with input address and output UTXOs
         {
             "amount":25,
             "inputs":[
                 "pay:null:1_3FvPC7dzFbQKzfG"
             ],
             "outputs":[
                 {"recipient":"pay:null:FrSVC3IrirScyRh","amount":5,"extra":null}
             ]
         }
        """
        if not hasattr(CredentialDef.get_payment_txn, "cb"):
            self.logger.debug("vcx_credentialdef_get_payment_txn: Creating callback")
            CredentialDef.get_payment_txn.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)

        payment_txn = await do_call('vcx_credentialdef_get_payment_txn',
                                    c_credential_handle,
                                    CredentialDef.get_payment_txn.cb)

        return json.loads(payment_txn.decode())


    async def update_state(self) -> int:
        """
        Checks if credential definition is published on the Ledger and updates the state.
        Possible states:
             0 = Built
             1 = Published
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        assert await credential_def.update_state() == PublicEntityState.Published
        :return: Current state of the credential definition
        """
        return await self._update_state(CredentialDef, 'vcx_credentialdef_update_state')


    async def get_state(self) -> int:
        """
        Get the current state of the credential definition object
        Possible states:
             0 = Built
             1 = Published
        Example:
        source_id = 'foobar123'
        schema_name = 'Schema Name'
        payment_handle = 0
        credential_def = await CredentialDef.create(source_id, name, schema_id, payment_handle)
        assert await credential_def.get_state() == PublicEntityState.Published
        :return:  Current state of the credential definition
        """
        return await self._get_state(CredentialDef, 'vcx_credentialdef_get_state')



from ctypes import CDLL
import sys
from vcx.logging import set_logger

LIBRARY = "vcx"


def _cdll() -> CDLL:
    if not hasattr(_cdll, "cdll"):
        _cdll.cdll = _load_cdll()
        set_logger(_cdll.cdll)

    return _cdll.cdll


def _load_cdll() -> CDLL:
    prefix_mapping = {"darwin": "lib", "linux": "lib", "linux2": "lib", "win32": ""}
    suffix_mapping = {"darwin": ".dylib", "linux": ".so", "linux2": ".so", "win32": ".dll"}

    os_name = sys.platform

    try:
        prefix = prefix_mapping[os_name]
        suffix = suffix_mapping[os_name]
    except KeyError:
        raise OSError("OS isn't supported: %s", os_name)

    library_name = "{0}{1}{2}".format(prefix, LIBRARY, suffix)

    try:
        res = CDLL(library_name)
        return res
    except OSError as e:
        raise e



from ctypes import *
import logging
from vcx.common import do_call, do_call_sync, create_cb

__all__ = ["vcx_init", "vcx_init_with_config"]

async def vcx_init(config_path: str) -> None:
    """
    Initializes VCX with config file.
    The list of available options see here: https://github.com/hyperledger/indy-sdk/blob/master/docs/configuration.md

    :param config_path: String - path to a config file to populate config attributes
    Example:
    await vcx_init('/home/username/vcxconfig.json')
    :return:
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_init, "cb"):
        logger.debug("vcx_init: Creating callback")
        vcx_init.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

    c_config_path = c_char_p(config_path.encode('utf-8'))

    result = await do_call('vcx_init',
                           c_config_path,
                           vcx_init.cb)

    logger.debug("vcx_init completed")
    return result


async def vcx_init_with_config(config: str) -> None:
    """
    Initializes VCX with config settings

    :param config: config as json.
    The list of available options see here: https://github.com/hyperledger/indy-sdk/blob/master/docs/configuration.md

    Example:
    config = {
      "agency_did": "VsKV7grR1BUE29mG2Fm2kX",
      "agency_verkey": "Hezce2UWMZ3wUhVkh2LfKSs8nDzWwzs2Win7EzNN3YaR",
      "agency_endpoint": "http://localhost:8080",
      "genesis_path":"/var/lib/indy/verity-staging/pool_transactions_genesis",
      "institution_name": "institution",
      "institution_logo_url": "http://robohash.org/234",
      "institution_did": "EwsFhWVoc3Fwqzrwe998aQ",
      "institution_verkey": "8brs38hPDkw5yhtzyk2tz7zkp8ijTyWnER165zDQbpK6",
      "remote_to_sdk_did": "EtfeMFytvYTKnWwqTScp9D",
      "remote_to_sdk_verkey": "8a7hZDyJK1nNCizRCKMr4H4QbDm8Gg2vcbDRab8SVfsi",
      "sdk_to_remote_did": "KacwZ2ndG6396KXJ9NDDw6",
      "sdk_to_remote_verkey": "B8LgZGxEPcpTJfZkeqXuKNLihM1Awm8yidqsNwYi5QGc"
    }
    await vcx_init_with_config(config)
    :return:
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_init_with_config, "cb"):
        logger.debug("vcx_init_with_config: Creating callback")
        vcx_init_with_config.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

    c_config = c_char_p(config.encode('utf-8'))

    result = await do_call('vcx_init_with_config',
                           c_config,
                           vcx_init_with_config.cb)

    logger.debug("vcx_init_with_config completed")
    return result

def vcx_init_minimal(config_string: str) -> None:
    """
    Initializes VCX with minimal (no-agency) config file AFTER the wallet and pool are set.

    :param config_string: String
    Example:
    vcx_wallet_set_handle(wallet_handle)
    vcx_pool_set_handle(pool_handle)
    await vcx_init_minimal('{"wallet_name":"wallet1",.....}')
    :return:
    """
    logger = logging.getLogger(__name__)

    c_config_string = c_char_p(config_string.encode('utf-8'))

    result = do_call_sync('vcx_init_minimal', c_config_string)

    logger.debug("vcx_init_minimal completed")
    return result


from ctypes import *

import logging
from logging import ERROR, WARNING, INFO, DEBUG, CRITICAL

TRACE = 5


def set_logger(cdll):
    logger = logging.getLogger(__name__)
    logging.addLevelName(TRACE, "TRACE")
    logging.basicConfig(level=CRITICAL)

    logger.debug("set_logger: >>>")

    def _log(context, level, target, message, module_path, file, line):
        libvcx_logger = logger.getChild('native.' + target.decode().replace('::', '.'))

        level_mapping = {1: ERROR, 2: WARNING, 3: INFO, 4: DEBUG, 5: TRACE, }

        libvcx_logger.log(level_mapping[level],
                          "\t%s:%d | %s",
                          file.decode(),
                          line,
                          message.decode())

    set_logger.callbacks = {
        'enabled_cb': None,
        'log_cb': CFUNCTYPE(None, c_void_p, c_int, c_char_p, c_char_p, c_char_p, c_char_p, c_int)(_log),
        'flush_cb': None
    }

    getattr(cdll, 'vcx_set_logger')(None,
                                    set_logger.callbacks['enabled_cb'],
                                    set_logger.callbacks['log_cb'],
                                    set_logger.callbacks['flush_cb'])

    logger.debug("set_logger: <<<")



from ctypes import *
from vcx.common import do_call, do_call_sync, create_cb
import json

import logging


class Wallet:

    @staticmethod
    async def close_search(handle: int):
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.close_search, "cb"):
            logger.debug("vcx_wallet_close_search: Creating callback")
            Wallet.close_search.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_handle = c_uint32(handle)

        data = await do_call('vcx_wallet_close_search',
                             c_handle,
                             Wallet.close_search.cb)

        logger.debug("vcx_wallet_close_search completed")
        return data

    @staticmethod
    async def open_search(type_: str, query: dict, options: dict):
        """
        Opens a search handle within the storage wallet.

        :param type_: String
        :param query: dictionary
        :param options: dictionary
        Example:
        query_json = {"tagName1": "str1"}
        type_ = 'TestType'
        search_handle = await Wallet.open_search(type_, query_json, None)
        :return: int
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.open_search, "cb"):
            logger.debug("vcx_wallet_open_search: Creating callback")
            Wallet.open_search.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_query = c_char_p(json.dumps(query).encode('utf-8'))
        c_options = c_char_p(json.dumps(options).encode('utf-8')) if options else None

        data = await do_call('vcx_wallet_open_search',
                             c_type_,
                             c_query,
                             c_options,
                             Wallet.open_search.cb)

        logger.debug("vcx_wallet_open_search completed")
        return data

    @staticmethod
    async def search_next_records(handle: int, count: int):
        """
        Searches for next n record from an open search handle

        :param handle: int
        :param count: int
         Example:
        query_json = {"tagName1": "str1"}
        type_ = 'TestType'
        search_handle = await Wallet.open_search(type_, query_json, None)
        results = await Wallet.search_next_records(search_handle, 5)
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.search_next_records, "cb"):
            logger.debug("vcx_wallet_search_next_records: Creating callback")
            Wallet.search_next_records.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))
        c_handle = c_uint32(handle)
        c_count = c_uint32(count)

        data = await do_call('vcx_wallet_search_next_records',
                             c_handle,
                             c_count,
                             Wallet.search_next_records.cb)

        logger.debug("vcx_wallet_search_next_records completed")
        return data.decode()

    @staticmethod
    async def get_record(type_: str, id: str, options: str):
        """
        Retrieves a record from the wallet storage.
        :param type_: String
        :param id: String
        :param options: String
        Example:
        import json
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': json.dumps({
                'tag1': 'unencrypted value1',
                '~encryptedTag', 'this value is encrypted,
                'integerTag', 1
                }),
            'type_': 'TestType',
            'value': 'RecordValue'
        })
        options = json.dumps({"retrieveType": True, "retrieveValue": True, "retrieveTags": True})
        record = await Wallet.get_record('TestType', 'RecordId', options)
        :return:
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.get_record, "cb"):
            logger.debug("vcx_wallet_get_record: Creating callback")
            Wallet.get_record.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_options = c_char_p(options.encode('utf-8'))
        data = await do_call('vcx_wallet_get_record',
                             c_type_,
                             c_id,
                             c_options,
                             Wallet.get_record.cb)

        logger.debug("vcx_wallet_get_record completed")
        return data.decode()

    @staticmethod
    async def delete_record(type_: str, id: str):
        """
        Delete a record from the storage wallet.

        :param type_:
        :param id:
        Example:
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': json.dumps({
                'tag1': 'unencrypted value1',
                '~encryptedTag', 'this value is encrypted,
                'integerTag', 1
                }),
            'type_': 'TestType',
            'value': 'RecordValue'
        })
        await Wallet.delete_record('TestType', 'RecordId')
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.delete_record, "cb"):
            logger.debug("vcx_wallet_delete_record: Creating callback")
            Wallet.delete_record.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        result = await do_call('vcx_wallet_delete_record',
                               c_type_,
                               c_id,
                               Wallet.delete_record.cb)

        logger.debug("vcx_wallet_delete_record completed")
        return result

    @staticmethod
    async def delete_record_tags(type_: str, id: str, tags: list ):
        """
        Delete tags associated with a record
        :param type_:
        :param id:
        :param tags:

        Example:
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': json.dumps({
                'tag1': 'unencrypted value1',
                '~encryptedTag', 'this value is encrypted,
                'integerTag', 1
                }),
            'type_': 'TestType',
            'value': 'RecordValue'
        })
        await Wallet.delete_record_tags('TestType', 'RecordId', ['tag1', 'integerTag'])
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.delete_record_tags, "cb"):
            logger.debug("vcx_wallet_delete_record_tags: Creating callback")
            Wallet.delete_record_tags.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        tags = json.dumps(tags)
        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_tags = c_char_p(tags.encode('utf-8'))
        result = await do_call('vcx_wallet_delete_record_tags',
                               c_type_,
                               c_id,
                               c_tags,
                               Wallet.delete_record_tags.cb)

        logger.debug("vcx_wallet_delete_record_tags completed")
        return result

    @staticmethod
    async def add_record_tags(type_: str, id: str, tags: str):
        """
        Adds tags to a record already stored in the storage wallet.
        :param type_: String
        :param id: String
        :param tags: String
        Example:
        import json
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': '{}',
            'type_': 'TestType',
            'value': 'RecordValue
        await Wallet.add_record_tags('TestType', 'RecordId', json.dumps({'addthistag':'valuetag1'}))
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.add_record_tags, "cb"):
            logger.debug("vcx_wallet_add_record_tags: Creating callback")
            Wallet.add_record_tags.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_tags = c_char_p(tags.encode('utf-8'))
        result = await do_call('vcx_wallet_add_record_tags',
                               c_type_,
                               c_id,
                               c_tags,
                               Wallet.add_record_tags.cb)

        logger.debug("vcx_wallet_add_record_tags completed")
        return result

    @staticmethod
    async def update_record_tags(type_: str, id: str, tags: str):
        """
        Updates the tags on a record, removing any previous value.
        :param type_: String
        :param id: String
        :param tags: String
        Example:
        import json
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': json.dumps({'foobar':'this value will get overwritten'}),
            'type_': 'TestType',
            'value': 'RecordValue
        await Wallet.update_record_tags('TestType', 'RecordId', json.dumps({'foobar':'new value'}))
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.update_record_tags, "cb"):
            logger.debug("vcx_wallet_update_record_tags: Creating callback")
            Wallet.update_record_tags.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_tags = c_char_p(tags.encode('utf-8'))
        result = await do_call('vcx_wallet_update_record_tags',
                               c_type_,
                               c_id,
                               c_tags,
                               Wallet.update_record_tags.cb)

        logger.debug("vcx_wallet_update_record_tags completed")
        return result

    @staticmethod
    async def update_record_value(type_: str, id: str, value: str):
        """
        Updates the value of a record
        :param type_: String
        :param id: String
        :param value: String
        Example:
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': '{}',
            'type_': 'TestType',
            'value': 'this will be overwritten' })
        await Wallet.update_record('TestType', 'RecordId', 'new value')
        :return: 
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.update_record_value, "cb"):
            logger.debug("vcx_wallet_update_record_value: Creating callback")
            Wallet.update_record_value.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_value = c_char_p(value.encode('utf-8'))
        result = await do_call('vcx_wallet_update_record_value',
                               c_type_,
                               c_id,
                               c_value,
                               Wallet.update_record_value.cb)

        logger.debug("vcx_wallet_update_record_value completed")
        return result

    @staticmethod
    async def add_record(type_: str, id: str, value: str, tags: str):
        """

        :param type_: String
        :param id: String
        :param value: String
        :param tags: Dictionary
        Example:
        import json
        await Wallet.add_record({
            'id': 'RecordId',
            'tags': json.dumps({
                'tag1': 'unencrypted value1',
                '~encryptedTag', 'this value is encrypted,
                'integerTag', 1
                }),
            'type_': 'TestType',
            'value': 'RecordValue'
        })
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.add_record, "cb"):
            logger.debug("vcx_wallet_add_record: Creating callback")
            Wallet.add_record.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_type_ = c_char_p(type_.encode('utf-8'))
        c_id = c_char_p(id.encode('utf-8'))
        c_value = c_char_p(value.encode('utf-8'))
        c_tags = c_char_p(tags.encode('utf-8'))
        result = await do_call('vcx_wallet_add_record',
                               c_type_,
                               c_id,
                               c_value,
                               c_tags,
                               Wallet.add_record.cb)

        logger.debug("vcx_wallet_add_record completed")
        return result

    @staticmethod
    async def get_token_info(handle: int) -> str:
        """
        Retrieves from the ledger token info associated with the wallet.
        :param handle:
        Example:
        payment_handle = 0 // payment handle is always 0, for now.
        info = await Wallet.get_token_info(payment_handle)
        :return: info
           "{"balance":6,"balance_str":"6","addresses":[{"address":"pay:null:9UFgyjuJxi1i1HD","balance":3,"utxo":[{"source":"pay:null:1","paymentAddress":"pay:null:zR3GN9lfbCVtHjp","amount":1,"extra":"yqeiv5SisTeUGkw"}]}]}"
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.get_token_info, "cb"):
            logger.debug("vcx_wallet_get_token_info: Creating callback")
            Wallet.get_token_info.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_payment = c_uint32(handle)

        result = await do_call('vcx_wallet_get_token_info',
                               c_payment,
                               Wallet.get_token_info.cb)

        logger.debug("vcx_wallet_get_token_info completed")
        return result

    @staticmethod
    async def create_payment_address(seed: str = None) -> str:
        """
        Creates a payment address inside the wallet.
        :param seed: String
        Example:
        address = await Wallet.create_payment_address('00000000000000000000000001234567')
        :return: String
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.create_payment_address, "cb"):
            logger.debug("vcx_wallet_create_payment_address: Creating callback")
            Wallet.create_payment_address.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        if seed:
            c_seed = c_char_p(seed.encode('utf-8'))
        else:
            c_seed = None

        result = await do_call('vcx_wallet_create_payment_address',
                               c_seed,
                               Wallet.create_payment_address.cb)

        logger.debug("vcx_wallet_create_payment_address completed")
        return result

    @staticmethod
    async def validate_payment_address(address: str) -> None:
        """
        Determines whether a payment address is valid or not
        :param address: String
        Example:
        address = await Wallet.create_payment_address('00000000000000000000000001234567')
        b = await Wallet.validate_payment_address(address)
        :return: Boolean
        """

        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.validate_payment_address, "cb"):
            logger.debug("vcx_wallet_validate_payment_address: Creating callback")
            Wallet.validate_payment_address.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_address = c_char_p(address.encode('utf-8'))
        result = await do_call('vcx_wallet_validate_payment_address',
                               c_address,
                               Wallet.validate_payment_address.cb)

        logger.debug("vcx_wallet_validate_payment_address completed")
        return result

    @staticmethod
    async def send_tokens(payment_handle: int, tokens: int, address: str) -> str:
        """
        Sends tokens to an address
        payment_handle is always 0
        :param payment_handle: Integer
        :param tokens: Integer
        :param address: String
        Example:
        payment_handle = 0
        amount = 1000
        address = await Wallet.create_payment_address('00000000000000000000000001234567')
        await Wallet.send_tokens(payment_handle, amount, address)
        :return:
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.send_tokens, "cb"):
            logger.debug("vcx_wallet_send_tokens: Creating callback")
            Wallet.send_tokens.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_payment_handle = c_uint32(payment_handle)
        c_tokens = c_char_p(str(tokens).encode('utf-8'))
        c_address = c_char_p(address.encode('utf-8'))

        result = await do_call('vcx_wallet_send_tokens',
                               c_payment_handle,
                               c_tokens,
                               c_address,
                               Wallet.send_tokens.cb)

        logger.debug("vcx_wallet_send_tokens completed")
        return result

    @staticmethod
    async def export(path, backup_key):
        """
        Exports opened wallet
        :param path: Path to export wallet to User's File System.
        :param backupKey: String representing the User's Key for securing (encrypting) the exported Wallet.
        :return:
        Error code - success indicates that the wallet was successfully exported.
        """
        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.export, "cb"):
            logger.debug("vcx_wallet_export: Creating callback")
            Wallet.export.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_backupKey = c_char_p(backup_key.encode('utf-8'))
        c_path = c_char_p(path.encode('utf-8'))

        result = await do_call('vcx_wallet_export',
                               c_path,
                               c_backupKey,
                               Wallet.export.cb)

        logger.debug("vcx_wallet_export completed")
        return result

    @staticmethod
    async def import_wallet(config):
        """
        Imports wallet from file with given key.
        Cannot be used if wallet is already opened (Especially if vcx_init has already been used).
        :param config: Can be same config that is passed to vcx_init.
        Must include: '{"wallet_name":"","wallet_key":"","exported_wallet_path":"","backup_key":""}'
        :return:
        Error code - success indicates that the wallet was successfully imported.
        """

        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.import_wallet, "cb"):
            logger.debug("vcx_wallet_import: Creating callback")
            Wallet.import_wallet.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_config = c_char_p(config.encode('utf-8'))

        result = await do_call('vcx_wallet_import',
                               c_config,
                               Wallet.import_wallet.cb)

        logger.debug("vcx_wallet_export completed")
        return result

    @staticmethod
    def set_handle(handle: int) -> None:
        """
        Sets the wallet handle for libvcx to use, called before vcx_init_minimal
        :param handle: wallet handle
        """
        c_handle = c_uint32(handle)

        do_call_sync('vcx_wallet_set_handle', c_handle)

    @staticmethod
    async def sign_with_address(address: str, msg: bytes) -> bytes:
        """
        Sign data using payment address
        :param payment_address
        :param msg:
        :return: signature
        """

        logger = logging.getLogger(__name__)

        def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32):
            return bytes(arr_ptr[:arr_len]),

        if not hasattr(Wallet.sign_with_address, "cb"):
            logger.debug("vcx_wallet_sign_with_address: Creating callback")
            Wallet.sign_with_address.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, POINTER(c_uint8), c_uint32), transform_cb)

        c_address = c_char_p(address.encode('utf-8'))
        c_msg_len = c_uint32(len(msg))

        result = await do_call('vcx_wallet_sign_with_address',
                               c_address,
                               msg,
                               c_msg_len,
                               Wallet.sign_with_address.cb)

        logger.debug("vcx_wallet_sign_with_address completed")
        return result

    @staticmethod
    async def verify_with_address(address: str, msg: bytes, signature: bytes) -> bool:
        """
        Verify the signature using payment address
        :param payment_address
        :param msg:
        :param signature:
        :return: bool
        """

        logger = logging.getLogger(__name__)

        if not hasattr(Wallet.verify_with_address, "cb"):
            logger.debug("vcx_wallet_verify_with_address: Creating callback")
            Wallet.verify_with_address.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_bool))

        c_address = c_char_p(address.encode('utf-8'))
        c_msg_len = c_uint32(len(msg))
        c_signature_len = c_uint32(len(signature))

        result = await do_call('vcx_wallet_verify_with_address',
                               c_address,
                               msg,
                               c_msg_len,
                               signature,
                               c_signature_len,
                               Wallet.verify_with_address.cb)

        logger.debug("vcx_wallet_sign_verify_address completed")
        return result



from typing import Optional
from ctypes import *
from vcx.common import do_call, create_cb
from vcx.error import VcxError, ErrorCode
from vcx.api.vcx_stateful import VcxStateful

import json


class Schema(VcxStateful):
    """
    Object that represents a schema written on the ledger.

    Attributes:
        source_id: user generated unique identifier
        schema_id: the ledger ID of the schema
        attrs: attribute/value pairs (the number of attributes should be less or equal than 125)
        version: version of the schema
        transaction: schema transaction that must be published to the ledger by Endorser
    """


    def __init__(self, source_id: str, name: str, version: str, attrs: list, transaction: Optional[str] = None):
        VcxStateful.__init__(self, source_id)
        self._source_id = source_id
        self._schema_id = None
        self._attrs = attrs
        self._name = name
        self._version = version
        self._transaction = transaction

    def __del__(self):
        self.release()
        self.logger.debug("Deleted {} obj: {}".format(Schema, self.handle))

    @property
    def schema_id(self):
        return self._schema_id

    @schema_id.setter
    def schema_id(self, x):
        self._schema_id = x

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, x):
        self._name = x

    @property
    def attrs(self):
        return self._attrs

    @attrs.setter
    def attrs(self, x):
        self._attrs = x

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, x):
        self._version = x


    @property
    def transaction(self):
        return self._transaction


    @transaction.setter
    def transaction(self, x):
        self._transaction = x


    @staticmethod
    async def create(source_id: str, name: str, version: str, attrs: list, payment_handle: int):
        """
        Creates a new schema object that is written to the ledger

        :param source_id: Institution's unique ID for the schema
        :param name: Name of schema
        :param version: Version of the schema
        :param attrs: Atttributes of the schema
        :param payment_handle: NYI - payment of ledger fee is taken from wallet automatically
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema = await Schema.create(source_id, name, version, attrs, payment_handle)
        :return: schema object, written to ledger
        """
        constructor_params = (source_id, name, version, attrs)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_name = c_char_p(name.encode('utf-8'))
        c_version = c_char_p(version.encode('utf-8'))
        c_schema_data = c_char_p(json.dumps(attrs).encode('utf-8'))
        c_payment = c_uint32(payment_handle)
        c_params = (c_source_id, c_name, c_version, c_schema_data, c_payment)

        schema = await Schema._create("vcx_schema_create", constructor_params, c_params)
        schema.schema_id = await schema.get_schema_id()
        return schema


    @staticmethod
    async def prepare_for_endorser(source_id: str, name: str, version: str, attrs: list, endorser: str):
        """
        Create a new Schema object that will be published by Endorser later.

        Note that CredentialDef can't be used for credential issuing until it will be published on the ledger.

        :param source_id: Institution's unique ID for the schema
        :param name: Name of schema
        :param version: Version of the schema
        :param attrs: Atttributes of the schema
        :param endorser: DID of the Endorser that will submit the transaction.
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        endorser = 'V4SGRU86Z58d6TV7PBUe6f'
        schema = await Schema.prepare_for_endorser(source_id, name, version, attrs, endorser)
        :return: schema object, schema transaction that should be passed to Endorser for witting to ledger
        """
        try:
            schema = Schema(source_id, '', '', [])

            if not hasattr(Schema.prepare_for_endorser, "cb"):
                schema.logger.debug("vcx_schema_prepare_for_endorser: Creating callback")
                Schema.prepare_for_endorser.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32, c_char_p))

            c_source_id = c_char_p(source_id.encode('utf-8'))
            c_name = c_char_p(name.encode('utf-8'))
            c_version = c_char_p(version.encode('utf-8'))
            c_schema_data = c_char_p(json.dumps(attrs).encode('utf-8'))
            c_endorser = c_char_p(endorser.encode('utf-8'))

            handle, transaction = await do_call('vcx_schema_prepare_for_endorser',
                                                c_source_id,
                                                c_name,
                                                c_version,
                                                c_schema_data,
                                                c_endorser,
                                                Schema.prepare_for_endorser.cb)
            schema.logger.debug("created schema object")

            schema.attrs = attrs
            schema.name = name
            schema.version = version
            schema.handle = handle
            schema.transaction = transaction
            return schema
        except KeyError:
            raise VcxError(ErrorCode.InvalidSchema)


    @staticmethod
    async def deserialize(data: dict):
        """
        Create the object from a previously serialized object.

        :param data: The output of the "serialize" call
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema1 = await Schema.create(source_id, name, version, attrs, payment_handle)
        data1 = await schema1.serialize()
        :return: A re-instantiated object
        """
        try:
            # Todo: Find better way to access attr_names. Potential for issues.
            schema = await Schema._deserialize("vcx_schema_deserialize",
                                               json.dumps(data),
                                               data['data']['source_id'],
                                               data['data']['name'],
                                               data['data']['version'],
                                               data['data']['data'])

            schema.schema_id = await schema.get_schema_id()
            return schema
        except KeyError:
            raise VcxError(ErrorCode.InvalidSchema)

    @staticmethod
    async def lookup(source_id: str, schema_id: str):
        """
        Create a new schema object from an existing ledger schema

        :param source_id: Institution's personal identification for the schema
        :param schema_id: Ledger schema ID for lookup
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema1 = await Schema.create(source_id, name, version, attrs, payment_handle)
        id1 = await schema.get_schema_id()
        data = await Schema.lookup(source_id, schema_id)
        assert data.attrs.sort() == ['sex', 'age', 'name', 'height'].sort()
        assert data.name == 'test-licence'
        assert data.handle > 0
        :return: schema object
        """
        try:
            schema = Schema(source_id, '', '', [])

            if not hasattr(Schema.lookup, "cb"):
                schema.logger.debug("vcx_schema_get_attributes: Creating callback")
                Schema.lookup.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32, c_char_p))

            c_source_id = c_char_p(source_id.encode('utf-8'))
            c_schema_id = c_char_p(schema_id.encode('utf-8'))

            handle, data = await do_call('vcx_schema_get_attributes',
                                         c_source_id,
                                         c_schema_id,
                                         Schema.lookup.cb)
            schema.logger.debug("created schema object")

            schema_result = json.loads(data.decode())
            schema.attrs = schema_result['data']
            schema.name = schema_result['name']
            schema.version = schema_result['version']
            schema.handle = handle
            return schema
        except KeyError:
            raise VcxError(ErrorCode.InvalidSchema)

    async def serialize(self) -> dict:
        """
        Serialize the object for storage

        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema1 = await Schema.create(source_id, name, version, attrs, payment_handle)
        data1 = await schema1.serialize()
        :return: serialized object
        """
        return await self._serialize(Schema, 'vcx_schema_serialize')

    def release(self) -> None:
        """
        destroy the object and release any memory associated with it

        :return: None
        """
        self._release(Schema, 'vcx_schema_release')

    async def get_schema_id(self):
        """
        Get the ledger ID of the object

        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema1 = await Schema.create(source_id, name, version, attrs, payment_handle)
        id1 = await schema.get_schema_id()
        :return: ID string
        """
        cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))
        c_handle = c_uint32(self.handle)
        id = await do_call('vcx_schema_get_schema_id', c_handle, cb)
        return id.decode()

    async def get_payment_txn(self):
        """
        Retrieve the txn associated with paying for the schema

        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema1 = await Schema.create(source_id, name, version, attrs, payment_handle)
        txn = await schema1.get_payment_txn()
        :return: JSON object with input address and output UTXOs
         {
             "amount":25,
             "inputs":[
                 "pay:null:1_3FvPC7dzFbQKzfG"
             ],
             "outputs":[
                 {"recipient":"pay:null:FrSVC3IrirScyRh","amount":5,"extra":null}
             ]
         }
        """
        if not hasattr(Schema.get_payment_txn, "cb"):
            self.logger.debug("vcx_schema_get_payment_txn: Creating callback")
            Schema.get_payment_txn.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)

        payment_txn = await do_call('vcx_schema_get_payment_txn',
                                    c_credential_handle,
                                    Schema.get_payment_txn.cb)

        return json.loads(payment_txn.decode())


    async def update_state(self) -> int:
        """
        Checks if schema is published on the Ledger and updates the state
        Possible states:
             0 = Built
             1 = Published
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema = await Schema.create(source_id, name, version, attrs, payment_handle)
        assert await schema.update_state() == PublicEntityState.Published
        :return: Current state of the schema
        """
        return await self._update_state(Schema, 'vcx_schema_update_state')


    async def get_state(self) -> int:
        """
        Get the current state of the schema object
        Possible states:
             0 = Built
             1 = Published
        Example:
        source_id = 'foobar123'
        name = 'Address Schema'
        version = '1.0'
        attrs = ['address', 'city', 'state']
        payment_handle = 0
        schema = await Schema.create(source_id, name, version, attrs, payment_handle)
        assert await schema.get_state() == PublicEntityState.Published
        :return:  Current internal state of the schema
        """
        return await self._get_state(Schema, 'vcx_schema_get_state')



from ctypes import *

import asyncio
import itertools
import logging
from .error import VcxError, ErrorCode, get_error_details
from vcx.cdll import _cdll

_futures = {}
_futures_counter = itertools.count()


def do_call(name: str, *args):
    logger = logging.getLogger(__name__)
    logger.debug("do_call: >>> name: %s, args: %s", name, args)

    event_loop = asyncio.get_event_loop()
    future = event_loop.create_future()
    command_handle = next(_futures_counter)

    _futures[command_handle] = (event_loop, future)

    err = getattr(_cdll(), name)(command_handle,
                                 *args)

    logger.debug("do_call: Function %s returned err: %i", name, err)

    if err != ErrorCode.Success:
        logger.warning("_do_call: Function %s returned error %i", name, err)
        error_details = get_error_details()
        future.set_exception(VcxError(ErrorCode(err), error_details))

    logger.debug("do_call: <<< %s", future)
    return future


def do_call_sync(name: str, *args):
    logger = logging.getLogger(__name__)
    logger.debug("do_call_sync: >>> name: %s, args: %s", name, args)

    err = getattr(_cdll(), name)(*args)

    logger.debug("do_call_sync: <<< %s", err)
    return err


def release(name, handle):
    logger = logging.getLogger(__name__)

    err = do_call_sync(name, handle)

    logger.debug("release: Function %s returned err: %i", name, err)

    if err != ErrorCode.Success:
        logger.warning("release: Function %s returned error %i", name, err)
        error_details = get_error_details()
        raise VcxError(ErrorCode(err), error_details)


def get_version() -> str:
    logger = logging.getLogger(__name__)

    name = 'vcx_version'
    c_version = do_call_sync(name)

    version = cast(c_version , c_char_p).value.decode()
    logger.debug("error_message: Function %s returned version: %s", name, version)

    return version


def update_institution_info(institution_name: str, logo_url: str) -> None:
    logger = logging.getLogger(__name__)

    name = 'vcx_update_institution_info'
    c_name = c_char_p(institution_name.encode('utf-8'))
    c_logo_url = c_char_p(logo_url.encode('utf-8'))

    do_call_sync(name, c_name, c_logo_url)
    logger.debug("vcx_init_with_config completed")


def shutdown(delete_wallet: bool):
    c_delete = c_bool(delete_wallet)
    name = 'vcx_shutdown'
    err = do_call_sync(name, c_delete)

    if err != ErrorCode.Success:
        error_details = get_error_details()
        raise VcxError(ErrorCode(err), error_details)


def mint_tokens():
    name = 'vcx_mint_tokens'
    do_call_sync(name, None, None)


def create_cb(cb_type: CFUNCTYPE, transform_fn=None):

    def _cb(command_handle: int, err: int, *args):
        if transform_fn:
            args = transform_fn(*args)
        error_details = get_error_details() if err != ErrorCode.Success else None
        error = VcxError(ErrorCode(err), error_details)
        _cxs_callback(command_handle, error, *args)

    res = cb_type(_cb)

    return res


def _cxs_callback(command_handle: int, err: VcxError, *args):
    (event_loop, future) = _futures[command_handle]
    event_loop.call_soon_threadsafe(_cxs_loop_callback, command_handle, err, *args)


def _cxs_loop_callback(command_handle: int, err: VcxError, *args):

    (event_loop, future) = _futures.pop(command_handle)

    if future.cancelled():
        print("_indy_loop_callback: Future was cancelled earlier")
    else:
        if err.error_code != ErrorCode.Success:
            future.set_exception(err)
        else:
            if len(args) == 0:
                res = None
            elif len(args) == 1:
                (res,) = args
            else:
                res = args

            future.set_result(res)




"""
The basic object of the VCX API.  Represents a pairwise relationship with another identity owner.  Once the
relationship, or connection, is established communication can happen securely and privately.  Credentials and
proofs are exchanged using this object.

TODO: document attributes
"""
from typing import Optional
from ctypes import *
from vcx.common import do_call, create_cb
from vcx.api.vcx_stateful import VcxStateful

import json


class Connection(VcxStateful):
    """
    The object of the VCX API representing a pairwise relationship with another identity owner.
    Once the relationship, or connection, is established communication can happen securely and privately.
    Credentials and Proofs are exchanged using this object.

    # States

    The set of object states and transitions depends on communication method is used.
    The communication method can be specified as config option on one of *_init function. The default communication method us `proprietary`.

    proprietary:
        Inviter:
            VcxStateType::VcxStateInitialized - once `vcx_connection_create` (create Connection object) is called.

            VcxStateType::VcxStateOfferSent - once `vcx_connection_connect` (send Connection invite) is called.

            VcxStateType::VcxStateAccepted - once `connReqAnswer` messages is received.
                                             use `vcx_connection_update_state` or `vcx_connection_update_state_with_message` functions for state updates.
            VcxStateType::VcxStateNone - once `vcx_connection_delete_connection` (delete Connection object) is called.

        Invitee:
            VcxStateType::VcxStateRequestReceived - once `vcx_connection_create_with_invite` (create Connection object with invite) is called.

            VcxStateType::VcxStateAccepted - once `vcx_connection_connect` (accept Connection invite) is called.

            VcxStateType::VcxStateNone - once `vcx_connection_delete_connection` (delete Connection object) is called.

    aries:
        Inviter:
            VcxStateType::VcxStateInitialized - once `vcx_connection_create` (create Connection object) is called.

            VcxStateType::VcxStateOfferSent - once `vcx_connection_connect` (prepared Connection invite) is called.

            VcxStateType::VcxStateRequestReceived - once `ConnectionRequest` messages is received.
                                                    accept `ConnectionRequest` and send `ConnectionResponse` message.
                                                    use `vcx_connection_update_state` or `vcx_connection_update_state_with_message` functions for state updates.

            VcxStateType::VcxStateAccepted - once `Ack` messages is received.
                                             use `vcx_connection_update_state` or `vcx_connection_update_state_with_message` functions for state updates.

            VcxStateType::VcxStateNone - once `vcx_connection_delete_connection` (delete Connection object) is called
                                            OR
                                        `ConnectionProblemReport` messages is received on state updates.

        Invitee:
            VcxStateType::VcxStateOfferSent - once `vcx_connection_create_with_invite` (create Connection object with invite) is called.

            VcxStateType::VcxStateRequestReceived - once `vcx_connection_connect` (accept `ConnectionInvite` and send `ConnectionRequest` message) is called.

            VcxStateType::VcxStateAccepted - once `ConnectionResponse` messages is received.
                                             send `Ack` message if requested.
                                             use `vcx_connection_update_state` or `vcx_connection_update_state_with_message` functions for state updates.

            VcxStateType::VcxStateNone - once `vcx_connection_delete_connection` (delete Connection object) is called
                                            OR
                                        `ConnectionProblemReport` messages is received on state updates.

    # Transitions

    proprietary:
        Inviter:
            VcxStateType::None - `vcx_connection_create` - VcxStateType::VcxStateInitialized
            VcxStateType::VcxStateInitialized - `vcx_connection_connect` - VcxStateType::VcxStateOfferSent
            VcxStateType::VcxStateOfferSent - received `connReqAnswer` - VcxStateType::VcxStateAccepted
            any state - `vcx_connection_delete_connection` - `VcxStateType::VcxStateNone`

        Invitee:
            VcxStateType::None - `vcx_connection_create_with_invite` - VcxStateType::VcxStateRequestReceived
            VcxStateType::VcxStateRequestReceived - `vcx_connection_connect` - VcxStateType::VcxStateAccepted
            any state - `vcx_connection_delete_connection` - `VcxStateType::VcxStateNone`

    aries - RFC: https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential
        Inviter:
            VcxStateType::None - `vcx_connection_create` - VcxStateType::VcxStateInitialized

            VcxStateType::VcxStateInitialized - `vcx_connection_connect` - VcxStateType::VcxStateOfferSent

            VcxStateType::VcxStateOfferSent - received `ConnectionRequest` - VcxStateType::VcxStateRequestReceived
            VcxStateType::VcxStateOfferSent - received `ConnectionProblemReport` - VcxStateType::VcxStateNone

            VcxStateType::VcxStateRequestReceived - received `Ack` - VcxStateType::VcxStateAccepted
            VcxStateType::VcxStateRequestReceived - received `ConnectionProblemReport` - VcxStateType::VcxStateNone

            VcxStateType::VcxStateAccepted - received `Ping`, `PingResponse`, `Query`, `Disclose` - VcxStateType::VcxStateAccepted

            any state - `vcx_connection_delete_connection` - VcxStateType::VcxStateNone


        Invitee:
            VcxStateType::None - `vcx_connection_create_with_invite` - VcxStateType::VcxStateOfferSent

            VcxStateType::VcxStateOfferSent - `vcx_connection_connect` - VcxStateType::VcxStateRequestReceived
            VcxStateType::VcxStateOfferSent - received `ConnectionProblemReport` - VcxStateType::VcxStateNone

            VcxStateType::VcxStateRequestReceived - received `ConnectionResponse` - VcxStateType::VcxStateAccepted
            VcxStateType::VcxStateRequestReceived - received `ConnectionProblemReport` - VcxStateType::VcxStateNone

            VcxStateType::VcxStateAccepted - received `Ping`, `PingResponse`, `Query`, `Disclose` - VcxStateType::VcxStateAccepted

            any state - `vcx_connection_delete_connection` - VcxStateType::VcxStateNone

    # Messages

    proprietary:
        ConnectionRequest (`connReq`)
        ConnectionRequestAnswer (`connReqAnswer`)

    aries:
        Invitation - https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#0-invitation-to-connect
        ConnectionRequest - https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#1-connection-request
        ConnectionResponse - https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#2-connection-response
        ConnectionProblemReport - https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#error-message-example
        Ack - https://github.com/hyperledger/aries-rfcs/tree/master/features/0015-acks#explicit-acks
        Ping - https://github.com/hyperledger/aries-rfcs/tree/master/features/0048-trust-ping#messages
        PingResponse - https://github.com/hyperledger/aries-rfcs/tree/master/features/0048-trust-ping#messages
        Query - https://github.com/hyperledger/aries-rfcs/tree/master/features/0031-discover-features#query-message-type
        Disclose - https://github.com/hyperledger/aries-rfcs/tree/master/features/0031-discover-features#disclose-message-type

    TODO: document attributes
    """

    def __init__(self, source_id: str):
        VcxStateful.__init__(self, source_id)

    def __del__(self):
        self.release()
        self.logger.debug("Deleted {} obj: {}".format(Connection, self.handle))

    @staticmethod
    async def create(source_id: str):
        """
        Create a connection object, represents a single endpoint and can be used for sending and receiving
        credentials and proofs

        :param source_id: Institution's unique ID for the connection
        :return: connection object
        Example:
        connection = await Connection.create(source_id)
        """
        constructor_params = (source_id,)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_params = (c_source_id,)

        return await Connection._create( "vcx_connection_create",
                                        constructor_params,
                                        c_params)

    @staticmethod
    async def create_with_details(source_id: str, invite_details: str):
        """
        Create a connection object with a provided invite, represents a single endpoint and can be used for sending and receiving
        credentials and proofs

        Invite details are provided by the entity offering a connection and generally pulled from a provided QRCode.
        :param source_id: Institution's unique ID for the connection
        :param invite_details: A string representing a json object which is provided by an entity that wishes to make a connection.
            Invite format depends on communication method:
                proprietary:
                    {"targetName": "", "statusMsg": "message created", "connReqId": "mugIkrWeMr", "statusCode": "MS-101", "threadId": null, "senderAgencyDetail": {"endpoint": "http://localhost:8080", "verKey": "key", "DID": "did"}, "senderDetail": {"agentKeyDlgProof": {"agentDID": "8f6gqnT13GGMNPWDa2TRQ7", "agentDelegatedKey": "5B3pGBYjDeZYSNk9CXvgoeAAACe2BeujaAkipEC7Yyd1", "signature": "TgGSvZ6+/SynT3VxAZDOMWNbHpdsSl8zlOfPlcfm87CjPTmC/7Cyteep7U3m9Gw6ilu8SOOW59YR1rft+D8ZDg=="}, "publicDID": "7YLxxEfHRiZkCMVNii1RCy", "name": "Faber", "logoUrl": "http://robohash.org/234", "verKey": "CoYZMV6GrWqoG9ybfH3npwH3FnWPcHmpWYUF8n172FUx", "DID": "Ney2FxHT4rdEyy6EDCCtxZ"}}
                aries: https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#0-invitation-to-connect
                 {
                    "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/connections/1.0/invitation",
                    "label": "Alice",
                    "recipientKeys": ["8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"],
                    "serviceEndpoint": "https://example.com/endpoint",
                    "routingKeys": ["8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"]
                 }

        Example:
        connection2 = await Connection.create_with_details('MyBank', invite_details)
        :return: connection object
        """
        constructor_params = (source_id,)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_invite_details = c_char_p(invite_details.encode('utf-8'))

        c_params = (c_source_id, c_invite_details, )

        return await Connection._create( "vcx_connection_create_with_invite",
                                        constructor_params,
                                        c_params)

    @staticmethod
    async def deserialize(data: dict):
        """
        Takes a json string representing a connection object and recreates an object matching the json.

        :param data: json string representing a connection object. Is an output of `serialize` function.
        Example:
        data = await connection1.serialize()
        connection2 = await Connection.deserialize(data)
        :return: A re-instantiated object
        """
        return await Connection._deserialize("vcx_connection_deserialize",
                                             json.dumps(data),
                                             data.get('source_id'))

    async def connect(self, options: str) -> str:
        """
        Connect securely and privately to the endpoint represented by the object.

        :param options: detailed connection options
        Example options:
        {"connection_type":"SMS","phone":"5555555555","use_public_did":true}
        or:
        {"connection_type":"QR"}
        Example code:
        connection = await Connection.create('Sally')
        invite_details = await connection.connect('{"connection_type":"QR"}')
        :return: the invite details sent via SMS or ready to be sent via some other mechanism (QR for example)
        """
        if not hasattr(Connection.connect, "cb"):
            self.logger.debug("vcx_connection_connect: Creating callback")
            Connection.connect.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)
        c_connection_data = c_char_p(options.encode('utf-8'))
        invite_details = await do_call('vcx_connection_connect',
                                       c_connection_handle,
                                       c_connection_data,
                                       Connection.connect.cb)
        return invite_details

    async def redirect(self, redirect_to) -> None:
        """
        Connect securely and privately to the endpoint represented by the object.

        :param redirect_to: Existing connection to redirect to

        Example code:
        connection = await Connection.create_with_details('Sally', invite_details)
        await connection.redirect(old_connection)
        """
        if not hasattr(Connection.redirect, "cb"):
            self.logger.debug("vcx_connection_redirect: Creating callback")
            Connection.redirect.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_connection_handle = c_uint32(self.handle)
        c_redirect_handle = c_uint32(redirect_to.handle)
        await do_call('vcx_connection_redirect',
                      c_connection_handle,
                      c_redirect_handle,
                      Connection.redirect.cb)

    async def get_redirect_details(self) -> str:
        if not hasattr(Connection.get_redirect_details, "cb"):
            self.logger.debug("vcx_connection_get_redirect_details: Creating callback")
            Connection.get_redirect_details.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)
        result = await do_call('vcx_connection_get_redirect_details',
                               c_connection_handle,
                               Connection.get_redirect_details.cb)

        self.logger.debug("vcx_connection_get_redirect_details completed")
        return json.loads(result.decode())

    async def send_message(self, msg: str, msg_type: str, msg_title: str, ref_msg_id: str = None) -> str:
        """
            Send a generic message to the connection
            :param msg: actual message to send
            :param msg_type: type of message to send. can be any string.
            :param msg_title: message title (user notification)
            :param ref_msg_id: if responding to a message, provide msg id

            Example options:
            msg: "HI" or "{"key": "value"}" or "{ "@type": "type", "@id": "518be002-de8e-456e-b3d5-8fe472477a86", "comment": "Hi. Are you listening?"}"
            msg_type: "Greeting"
            msg_title: "Hi There"

            Example code:
            connection = await Connection.create_with_details('MyBank', invite_details)
            await connection.send_message("HI", "Greeting", "Hi There")

            :return: response message
            """
        if not hasattr(Connection.send_message, "cb"):
            self.logger.debug("vcx_connection_send_message: Creating callback")
            Connection.send_message.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        send_msg_options = {
            "msg_type": msg_type,
            "msg_title": msg_title,
            "ref_msg_id": ref_msg_id
        }
        c_connection_handle = c_uint32(self.handle)
        c_msg = c_char_p(msg.encode('utf-8'))
        c_send_msg_options = c_char_p(json.dumps(send_msg_options).encode('utf-8'))

        result = await do_call('vcx_connection_send_message',
                               c_connection_handle,
                               c_msg,
                               c_send_msg_options,
                               Connection.send_message.cb)

        self.logger.debug("vcx_connection_send_message completed")
        return result

    async def sign_data(self, msg: bytes) -> bytes:
        """
        Sign data using connection's pairwise key
        :param msg: message to sign represented as bytes
        :return: signature
        """

        def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32):
            return bytes(arr_ptr[:arr_len]),

        if not hasattr(Connection.sign_data, "cb"):
            self.logger.debug("vcx_connection_sign_data: Creating callback")
            Connection.sign_data.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, POINTER(c_uint8), c_uint32), transform_cb)

        c_connection_handle = c_uint32(self.handle)
        c_msg_len = c_uint32(len(msg))

        result = await do_call('vcx_connection_sign_data',
                               c_connection_handle,
                               msg,
                               c_msg_len,
                               Connection.sign_data.cb)

        self.logger.debug("vcx_connection_sign_data completed")
        return result

    async def verify_signature(self, msg: bytes, signature: bytes) -> bool:
        """
        Verify the signature is valid for the specified data using connection pairwise keys
        :param msg: message was signed
        :param signature: generated signature
        :return: bool - whether the signature was valid or not
        """
        if not hasattr(Connection.verify_signature, "cb"):
            self.logger.debug("vcx_connection_verify_signature: Creating callback")
            Connection.verify_signature.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_bool))

        c_connection_handle = c_uint32(self.handle)
        c_msg_len = c_uint32(len(msg))
        c_signature_len = c_uint32(len(signature))

        result = await do_call('vcx_connection_verify_signature',
                               c_connection_handle,
                               msg,
                               c_msg_len,
                               signature,
                               c_signature_len,
                               Connection.verify_signature.cb)

        self.logger.debug("vcx_connection_verify_signature completed")
        return result

    async def _delete(self):
        if not hasattr(Connection._delete, "cb"):
            Connection._delete.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_connection_handle = c_uint32(self.handle)
        return await do_call('vcx_connection_delete_connection', c_connection_handle, Connection._delete.cb)

    async def serialize(self) -> dict:
        """
        Takes the Connection object and returns a json string of all its attributes
        Example:
        data = await connection1.serialize()
        :return: serialized object
        """
        return await self._serialize(Connection, 'vcx_connection_serialize')

    async def update_state(self) -> int:
        """
        Query the agency for the received messages.
        Checks for any messages changing state in the connection object and updates the state attribute.
        Example:
        connection = await Connection.create(source_id)
        assert await connection.update_state() == State.Initialized
        :return: Current state of the connection. Possible states:
                                                     1 - Initialized
                                                     2 - Request Sent
                                                     3 - Offer Received
                                                     4 - Accepted
        """
        return await self._update_state(Connection, 'vcx_connection_update_state')

    async def update_state_with_message(self, message: str) -> int:
        """
        Update the state of the connection based on the given message.
        :param msg: message to process
        Example:
        connection = await Connection.create(source_id)
        assert await connection.update_state_with_message(message) == State.Accepted
        :param message: message to process for state changes
        :return Current state of the connection. Possible states:
                                                    1 - Initialized
                                                    2 - Request Sent
                                                    3 - Offer Received
                                                    4 - Accepted
        """
        return await self._update_state_with_message(Connection, message, 'vcx_connection_update_state_with_message')

    async def get_state(self) -> int:
        """
        Returns the current internal state of the connection. Does NOT query agency for state updates.
        Possible states:
            1 - Initialized
            2 - Offer Sent
            3 - Request Received
            4 - Accepted
        Example:
        connection = await Connection.create(source_id)
        assert await connection.get_state() == State.Initialized
        :return:  Current internal state of the connection
        """
        return await self._get_state(Connection, 'vcx_connection_get_state')

    def release(self) -> None:
        """
        destroy the object and release any memory associated with it

        :return: None
        """
        self._release(Connection, 'vcx_connection_release')

    async def delete(self):
        """
        Delete the object from the agency and release any memory associated with it.

        NOTE: This eliminates the connection and any ability to use it for any communication.

        Example:
        connection = await Connection.create(source_id)
        await connection.delete()
        :return: None
        """
        await self._delete()

    async def invite_details(self, abbreviated: bool) -> dict:
        """
        Get the invite details that were sent or can be sent to the remote side.

        :param abbreviated: abbreviate invite details or not (applicable for `proprietary` communication method only)
        Example:
        phone_number = '8019119191'
        connection = await Connection.create('foobar123')
        invite_details = await connection.connect(phone_number)
        inivte_details_again = await connection.invite_details()
        :return: JSON of invite_details sent to connection
            Invite format depends on communication method:
                proprietary:
                    {"targetName": "", "statusMsg": "message created", "connReqId": "mugIkrWeMr", "statusCode": "MS-101", "threadId": null, "senderAgencyDetail": {"endpoint": "http://localhost:8080", "verKey": "key", "DID": "did"}, "senderDetail": {"agentKeyDlgProof": {"agentDID": "8f6gqnT13GGMNPWDa2TRQ7", "agentDelegatedKey": "5B3pGBYjDeZYSNk9CXvgoeAAACe2BeujaAkipEC7Yyd1", "signature": "TgGSvZ6+/SynT3VxAZDOMWNbHpdsSl8zlOfPlcfm87CjPTmC/7Cyteep7U3m9Gw6ilu8SOOW59YR1rft+D8ZDg=="}, "publicDID": "7YLxxEfHRiZkCMVNii1RCy", "name": "Faber", "logoUrl": "http://robohash.org/234", "verKey": "CoYZMV6GrWqoG9ybfH3npwH3FnWPcHmpWYUF8n172FUx", "DID": "Ney2FxHT4rdEyy6EDCCtxZ"}}
                aries: https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol#0-invitation-to-connect
                 {
                    "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/connections/1.0/invitation",
                    "label": "Alice",
                    "recipientKeys": ["8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"],
                    "serviceEndpoint": "https://example.com/endpoint",
                    "routingKeys": ["8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"]
                 }

        """
        if not hasattr(Connection.invite_details, "cb"):
            self.logger.debug("vcx_connection_invite_details: Creating callback")
            Connection.invite_details.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)
        c_abbreviated = c_bool(abbreviated)

        details = await do_call('vcx_connection_invite_details',
                                c_connection_handle,
                                c_abbreviated,
                                Connection.invite_details.cb)

        return json.loads(details.decode())


    async def send_ping(self, comment: Optional[str] = None):
        """
        Send trust ping message to the specified connection to prove that two agents have a functional pairwise channel.

        Note that this function is useful in case `aries` communication method is used.
        In other cases it returns ActionNotSupported error.
        :param comment: (Optional) human-friendly description of the ping.

        :return: no value
        """
        if not hasattr(Connection.send_ping, "cb"):
            self.logger.debug("vcx_connection_send_ping: Creating callback")
            Connection.send_ping.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_connection_handle = c_uint32(self.handle)
        c_comment = c_char_p(comment.encode('utf-8')) if comment is not None else None

        await do_call('vcx_connection_send_ping',
                      c_connection_handle,
                      c_comment,
                      Connection.send_ping.cb)


    async def send_discovery_features(self, query: Optional[str] = None, comment: Optional[str] = None):
        """
        Send discovery features message to the specified connection to discover which features it supports, and to what extent.

        Note that this function is useful in case `aries` communication method is used.
        In other cases it returns ActionNotSupported error.

        :param query: (Optional) query string to match against supported message types.
        :param comment: (Optional) human-friendly description of the ping.
        :return: no value
        """
        if not hasattr(Connection.send_discovery_features, "cb"):
            self.logger.debug("vcx_connection_send_discovery_features: Creating callback")
            Connection.send_discovery_features.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_connection_handle = c_uint32(self.handle)
        c_query = c_char_p(query.encode('utf-8')) if query is not None else None
        c_comment = c_char_p(comment.encode('utf-8')) if comment is not None else None

        await do_call('vcx_connection_send_discovery_features',
                      c_connection_handle,
                      c_query,
                      c_comment,
                      Connection.send_discovery_features.cb)

    async def get_my_pw_did(self) -> str:
        if not hasattr(Connection.get_my_pw_did, "cb"):
            self.logger.debug("get_my_pw_did: Creating callback")
            Connection.get_my_pw_did.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)

        my_pw_did = await do_call('vcx_connection_get_pw_did', c_connection_handle, Connection.get_my_pw_did.cb)
        return my_pw_did.decode()

    async def get_their_pw_did(self) -> str:
        if not hasattr(Connection.get_their_pw_did, "cb"):
            self.logger.debug("get_their_pw_did: Creating callback")
            Connection.get_their_pw_did.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)

        their_pw_did =\
            await do_call('vcx_connection_get_their_pw_did', c_connection_handle, Connection.get_their_pw_did.cb)
        return their_pw_did.decode()

    async def info(self) -> dict:
        """
        Get the information regarding the connection object.
        Note: This method can be used for `aries` communication method only.
            For other communication method it returns ActionNotSupported error.
        :return: JSON of connection information
        {
            "current": {
                "did": <str>
                "recipientKeys": array<str>
                "routingKeys": array<str>
                "serviceEndpoint": <str>,
                "protocols": array<str> -  The set of protocol supported by current side.
            },
            "remote: { <Option> - details about remote connection side
                "did": <str> - DID of remote side
                "recipientKeys": array<str> - Recipient keys
                "routingKeys": array<str> - Routing keys
                "serviceEndpoint": <str> - Endpoint
                "protocols": array<str> -
                    The set of protocol supported by side. Is filled after DiscoveryFeatures process was completed.
             }
        }
        """
        if not hasattr(Connection.info, "cb"):
            self.logger.debug("vcx_connection_info: Creating callback")
            Connection.info.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(self.handle)

        details = await do_call('vcx_connection_info',
                                c_connection_handle,
                                Connection.info.cb)

        return json.loads(details.decode())


import json
from ctypes import *
import logging
from typing import Optional

from vcx.common import do_call, create_cb, do_call_sync


async def vcx_agent_provision(config: str) -> None:
    """
    Provision an agent in the agency, populate configuration and wallet for this agent.
    Example:
    import json
    enterprise_config = {
        'agency_url': 'http://localhost:8080',
        'agency_did': 'VsKV7grR1BUE29mG2Fm2kX',
        'agency_verkey': "Hezce2UWMZ3wUhVkh2LfKSs8nDzWwzs2Win7EzNN3YaR",
        'wallet_name': 'LIBVCX_SDK_WALLET',
        'agent_seed': '00000000000000000000000001234561',
        'enterprise_seed': '000000000000000000000000Trustee1',
        'wallet_key': '1234'
    }
    vcx_config = await vcx_agent_provision(json.dumps(enterprise_config))
    :param config: JSON configuration
    :return: Configuration for vcx_init call.
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_agent_provision, "cb"):
        logger.debug("vcx_agent_provision: Creating callback")
        vcx_agent_provision.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

    c_config = c_char_p(config.encode('utf-8'))

    result = await do_call('vcx_agent_provision_async',
                           c_config,
                           vcx_agent_provision.cb)

    logger.debug("vcx_agent_provision completed")
    return result.decode()


async def vcx_agent_update_info(config: str) -> None:
    """
    Update information on the agent (ie, comm method and type)
    :param config:
    :return:
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_agent_update_info, "cb"):
        logger.debug("vcx_agent_update_info: Creating callback")
        vcx_agent_update_info.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

    c_config = c_char_p(config.encode('utf-8'))

    result = await do_call('vcx_agent_update_info',
                           c_config,
                           vcx_agent_update_info.cb)

    logger.debug("vcx_agent_update_info completed")
    return result


async def vcx_ledger_get_fees() -> str:
    """
    Get ledger fees from the sovrin network
    Example:
    fees = await vcx_ledger_get_fees()
    :return: JSON representing fees
        { "txnType1": amount1, "txnType2": amount2, ..., "txnTypeN": amountN }
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_ledger_get_fees, "cb"):
        logger.debug("vcx_ledger_get_fees: Creating callback")
        vcx_ledger_get_fees.cb = create_cb(CFUNCTYPE(None, c_uint32))

    result = await do_call('vcx_ledger_get_fees',
                           vcx_ledger_get_fees.cb)

    logger.debug("vcx_ledger_get_fees completed")
    return result


async def vcx_messages_download(status: str = None, uids: str = None, pw_dids: str = None) -> str:
    """
    Retrieve messages from the agent
    :param status: optional, comma separated - query for messages with the specified status.
                                     Possible statuses:
                                     MS-101 - Created
                                     MS-102 - Sent
                                     MS-103 - Received
                                     MS-104 - Accepted
                                     MS-105 - Rejected
                                     MS-106 - Reviewed
    :param uids: optional, comma separated - query for messages with the specified uids
    :param pw_dids: optional, comma separated - DID's pointing to specific connection
    :return: message
        [{"pairwiseDID":"did","msgs":[{"statusCode":"MS-106","payload":null,"senderDID":"","uid":"6BDkgc3z0E","type":"aries","refMsgId":null,"deliveryDetails":[],"decryptedPayload":"{"@msg":".....","@type":{"fmt":"json","name":"aries","ver":"1.0"}}"}]}]
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_messages_download, "cb"):
        logger.debug("vcx_messages_download: Creating callback")
        vcx_messages_download.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

    if status:
        c_status = c_char_p(status.encode('utf-8'))
    else:
        c_status = None

    if uids:
        c_uids = c_char_p(uids.encode('utf-8'))
    else:
        c_uids = None

    if pw_dids:
        c_pw_dids = c_char_p(pw_dids.encode('utf-8'))
    else:
        c_pw_dids = None

    result = await do_call('vcx_messages_download',
                           c_status,
                           c_uids,
                           c_pw_dids,
                           vcx_messages_download.cb)

    logger.debug("vcx_messages_download completed")
    return result


async def vcx_messages_update_status(msg_json: str):
    """
    Update the status of messages from the specified connection
    :param msg_json: messages to update:
        [
            {
                "pairwiseDID":"string",
                "uids":["string"]
            },
            ...
        ]
    :return:
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_messages_update_status, "cb"):
        logger.debug("vcx_messages_update_status: Creating callback")
        vcx_messages_update_status.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

    c_msg_json = c_char_p(msg_json.encode('utf-8'))
    c_status = c_char_p("MS-106".encode('utf-8'))

    result = await do_call('vcx_messages_update_status',
                           c_status,
                           c_msg_json,
                           vcx_messages_update_status.cb)

    logger.debug("vcx_messages_update_status completed")
    return result


def vcx_pool_set_handle(handle: int) -> None:
    """
    Sets the pool handle for libvcx to use, called before vcx_init_minimal
    :param handle: pool handle
    """
    c_handle = c_uint32(handle)

    do_call_sync('vcx_pool_set_handle', c_handle)


async def vcx_get_ledger_author_agreement():
    """
    Retrieve author agreement and acceptance mechanisms set on the Ledger
    :return: {"text":"Default agreement", "version":"1.0.0", "aml": {"label1": "description"}}
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_get_ledger_author_agreement, "cb"):
        logger.debug("vcx_get_ledger_author_agreement: Creating callback")
        vcx_get_ledger_author_agreement.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))
    result = await do_call('vcx_get_ledger_author_agreement',
                           vcx_get_ledger_author_agreement.cb)

    logger.debug("vcx_get_ledger_author_agreement completed")
    return result.decode()


def vcx_set_active_txn_author_agreement_meta(text: Optional[str],
                                             version: Optional[str],
                                             hash: Optional[str],
                                             acc_mech_type: str,
                                             time_of_acceptance: int) -> None:
    """
    Set some accepted agreement as active.
    As result of successful call of this function appropriate metadata will be appended to each write request.
    
    :param text and version - (optional) raw data about TAA from ledger.
               These parameters should be passed together.
               These parameters are required if hash parameter is ommited.
    :param hash - (optional) hash on text and version. This parameter is required if text and version parameters are ommited.
    :param acc_mech_type - mechanism how user has accepted the TAA
    :param time_of_acceptance - UTC timestamp when user has accepted the TAA

    :return: no value
    """
    logger = logging.getLogger(__name__)

    name = 'vcx_set_active_txn_author_agreement_meta'

    c_text = c_char_p(json.dumps(text).encode('utf-8')) if text else None
    c_version = c_char_p(json.dumps(version).encode('utf-8')) if version else None
    c_hash = c_char_p(json.dumps(hash).encode('utf-8')) if hash else None
    c_acc_mech_type = c_char_p(acc_mech_type.encode('utf-8'))
    c_time_of_acceptance = c_uint64(time_of_acceptance)

    do_call_sync(name, c_text, c_version, c_hash, c_acc_mech_type, c_time_of_acceptance)
    logger.debug("set_active_txn_author_agreement_meta completed")


async def vcx_get_request_price(action_json: str,
                                requester_info_json: Optional[str]):
    """
    Update the status of messages from the specified connection
    :param action_json: {
         "auth_type": ledger transaction alias or associated value,
         "auth_action": type of an action.,
         "field": transaction field,
         "old_value": (Optional) old value of a field, which can be changed to a new_value (mandatory for EDIT action),
         "new_value": (Optional) new value that can be used to fill the field,
     }
    :param requester_info_json: (Optional) {
         "role": string - role of a user which can sign transaction.
         "count": string - count of users.
         "is_owner": bool - if user is an owner of transaction.
     } otherwise context info will be used

    :return: price - tokens amount required for action performing
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_get_request_price, "cb"):
        logger.debug("vcx_get_request_price: Creating callback")
        vcx_get_request_price.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint64))

    c_action_json = c_char_p(action_json.encode('utf-8'))
    c_requester_info_json = c_char_p(requester_info_json.encode('utf-8')) if requester_info_json is not None else None

    result = await do_call('vcx_get_request_price',
                           c_action_json,
                           c_requester_info_json,
                           vcx_get_request_price.cb)

    logger.debug("vcx_get_request_price completed")
    return result


async def vcx_endorse_transaction(transaction: str) -> None:
    """
    Endorse transaction to the ledger preserving an original author
    :param transaction: transaction to endorse
    :return:
    """
    logger = logging.getLogger(__name__)

    if not hasattr(vcx_endorse_transaction, "cb"):
        logger.debug("vcx_endorse_transaction: Creating callback")
        vcx_endorse_transaction.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

    c_transaction = c_char_p(transaction.encode('utf-8'))

    result = await do_call('vcx_endorse_transaction',
                           c_transaction,
                           vcx_endorse_transaction.cb)

    logger.debug("vcx_endorse_transaction completed")
    return result




from enum import IntEnum


class State(IntEnum):
    Undefined = 0,
    Initialized = 1,
    OfferSent = 2,
    RequestReceived = 3,
    Accepted = 4,
    Unfulfilled = 5,
    Expired = 6,
    Revoked = 7,
    Redirected = 8,
    Rejected = 9,


class ProofState(IntEnum):
    Undefined = 0,
    Verified = 1,
    Invalid = 2


class PublicEntityState(IntEnum):
    Built = 0,
    Published = 1



from ctypes import *
from vcx.common import do_call, create_cb
from vcx.api.connection import Connection
from vcx.api.vcx_stateful import VcxStateful

import json


class Credential(VcxStateful):
    """
    The object of the VCX API representing a Holder side in the credential issuance process.
    Assumes that pairwise connection between Issuer and Holder is already established.

    # State

    The set of object states and transitions depends on communication method is used.
    The communication method can be specified as config option on one of *_init function. The default communication method us `proprietary`.

    proprietary:
        VcxStateType::VcxStateRequestReceived - once `vcx_credential_create_with_offer` (create Credential object) is called.

        VcxStateType::VcxStateOfferSent - once `vcx_credential_send_request` (send `CRED_REQ` message) is called.

        VcxStateType::VcxStateAccepted - once `CRED` messages is received.
                                         use `vcx_credential_update_state` or `vcx_credential_update_state_with_message` functions for state updates.

    aries:
        VcxStateType::VcxStateRequestReceived - once `vcx_credential_create_with_offer` (create Credential object) is called.

        VcxStateType::VcxStateOfferSent - once `vcx_credential_send_request` (send `CredentialRequest` message) is called.

        VcxStateType::VcxStateAccepted - once `Credential` messages is received.
        VcxStateType::None - once `ProblemReport` messages is received.
                                                use `vcx_credential_update_state` or `vcx_credential_update_state_with_message` functions for state updates.

    # Transitions

    proprietary:
        VcxStateType::None - `vcx_credential_create_with_offer` - VcxStateType::VcxStateRequestReceived

        VcxStateType::VcxStateRequestReceived - `vcx_credential_send_request` - VcxStateType::VcxStateOfferSent

        VcxStateType::VcxStateOfferSent - received `CRED` - VcxStateType::VcxStateAccepted

    aries: RFC - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential
        VcxStateType::None - `vcx_credential_create_with_offer` - VcxStateType::VcxStateRequestReceived

        VcxStateType::VcxStateRequestReceived - `vcx_issuer_send_credential_offer` - VcxStateType::VcxStateOfferSent

        VcxStateType::VcxStateOfferSent - received `Credential` - VcxStateType::VcxStateAccepted
        VcxStateType::VcxStateOfferSent - received `ProblemReport` - VcxStateType::None

    # Messages

    proprietary:
        CredentialOffer (`CRED_OFFER`)
        CredentialRequest (`CRED_REQ`)
        Credential (`CRED`)

    aries:
        CredentialProposal - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#propose-credential
        CredentialOffer - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#offer-credential
        CredentialRequest - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#request-credential
        Credential - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0036-issue-credential#issue-credential
        ProblemReport - https://github.com/hyperledger/aries-rfcs/tree/7b6b93acbaf9611d3c892c4bada142fe2613de6e/features/0035-report-problem#the-problem-report-message-type
        Ack - https://github.com/hyperledger/aries-rfcs/tree/master/features/0015-acks#explicit-acks
    """

    def __init__(self, source_id: str):
        VcxStateful.__init__(self, source_id)
        self._name = source_id
        self._cred_offer = None

    def __del__(self):
        self.release()
        self.logger.debug("Deleted {} obj: {}".format(Credential, self.handle))

    @property
    def cred_offer(self):
        return self._cred_offer

    @cred_offer.setter
    def cred_offer(self, x):
        self._cred_offer = x

    @staticmethod
    async def create(source_id: str, credential_offer: str):
        """
        Creates a credential with an offer.
        :param source_id: user defined id of object.
        :param credential_offer: JSON string representing the offer used as the basis of creation.
        :return: A created credential
        Example:
        offer depends on communication method:
         proprietary:
            [{
               "msg_type": "CLAIM_OFFER",
               "version": "0.1",
               "to_did": "8XFh8yBzrpJQmNyZzgoTqB",
               "from_did": "8XFh8yBzrpJQmNyZzgoTqB",
               "libindy_offer": '{}',
               "credential_attrs": {
                  "address1": [
                     "101 Tela Lane"
                  ],
                  "address2": [
                     "101 Wilson Lane"
                  ],
                  "city": [
                     "SLC"
                  ],
                  "state": [
                     "UT"
                  ],
                  "zip": [
                     "87121"
                  ]
               },
               "schema_seq_no": 1487,
               "cred_def_id": "id1",
               "claim_name": "Credential",
               "claim_id": "defaultCredentialId",
               "msg_ref_id": None,
            }]
          aries:
            {
                "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/issue-credential/1.0/offer-credential",
                "@id": "<uuid-of-offer-message>",
                "comment": "some comment",
                "credential_preview": <json-ld object>,
                "offers~attach": [
                    {
                        "@id": "libindy-cred-offer-0",
                        "mime-type": "application/json",
                        "data": {
                            "base64": "<bytes for base64>"
                        }
                    }
                ]
            }
        credential = await Credential.create(source_id, offer)
        """
        constructor_params = (source_id,)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_offer = c_char_p(json.dumps(credential_offer).encode('utf-8'))
        c_params = (c_source_id, c_offer, )

        return await Credential._create("vcx_credential_create_with_offer",
                                        constructor_params,
                                        c_params)

    @staticmethod
    async def create_with_msgid(source_id: str, connection: Connection, msg_id: str):
        """
        Create a credential based off of a known message id for a given connection.
        :param source_id: user defined id of object.
        :param connection: connection to receive offer from
        :param msg_id: id of the message that contains the credential offer
        :return: A created credential
        Example:
        credential = await Credential.create_with_msgid(source_id, connection, msg_id)
        assert await credential.get_state() == State.RequestReceived
        """
        credential = Credential(source_id,)

        c_source_id = c_char_p(source_id.encode('utf-8'))
        c_msg_id = c_char_p(json.dumps(msg_id).encode('utf-8'))
        c_connection_handle = c_uint32(connection.handle)

        if not hasattr(Credential.create_with_msgid, "cb"):
            Credential.create_with_msgid.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32, c_char_p))

        credential.handle, cred_offer = await do_call('vcx_credential_create_with_msgid',
                                                      c_source_id,
                                                      c_connection_handle,
                                                      c_msg_id,
                                                      Credential.create_with_msgid.cb)

        credential.cred_offer = json.loads(cred_offer.decode())

        return credential

    @staticmethod
    async def deserialize(data: dict):
        """
        Create a credential object from a previously serialized credential object
        :param data: JSON data from a serialized object.
        :return: A created credential
        Example:
        credential = await Credential.create(source_id, offer)
        data = await credential.serialize()
        credential2 = await Credential.deserialize(data)
        """

        credential = await Credential._deserialize("vcx_credential_deserialize",
                                                   json.dumps(data),
                                                   data.get('data').get('source_id'))
        return credential

    @staticmethod
    async def get_offers(connection: Connection) -> dict:
        """
        Retrieves all pending credential offers for a given connection.
        :param connection: A connection to query for credential offers.
        :return: A list of dictionary objects representing offers from a given connection.
         "[[{"msg_type": "CREDENTIAL_OFFER","version": "0.1","to_did": "...","from_did":"...","credential": {"account_num": ["...."],"name_on_account": ["Alice"]},"schema_seq_no": 48,"issuer_did": "...","credential_name": "Account Certificate","credential_id": "3675417066","msg_ref_id": "ymy5nth"}]]"
        Example:
        credential = await Credential.create_with_msgid(source_id, connection, msg_id)
        offers = await credential.get_offers(connection)
        """
        if not hasattr(Credential.get_offers, "cb"):
            Credential.get_offers.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_connection_handle = c_uint32(connection.handle)

        data = await do_call('vcx_credential_get_offers',
                             c_connection_handle,
                             Credential.get_offers.cb)

        return json.loads(data.decode())

    async def serialize(self) -> dict:
        """
        Serializes the crednetial object for storage and later deserialization.
        :return: dict representing the object
        Example:
        credential = await Credential.create_with_msgid(source_id, connection, msg_id)
        data = credential.serialzie()
        """
        return await self._serialize(Credential, 'vcx_credential_serialize')

    async def update_state(self) -> int:
        """
        Query the agency for the received messages.
        Checks for any messages changing state in the credential object and updates the state attribute.
        If it detects a credential it will store the credential in the wallet.

        Example:
        credential = await Credential.create(source_id, offer)
        credential.update_state()
        :return:
        """
        return await self._update_state(Credential, 'vcx_credential_update_state')

    async def update_state_with_message(self, message: str) -> int:

        """
        Update the state of the credential based on the given message.
        Example:
        proof = await Proof.create(source_id)
        assert await proof.update_state_with_message(message) == State.Accepted
        :param message: message to process for state changes
        :return Current state of the credential
        """
        return await self._update_state_with_message(Credential, message, 'vcx_credential_update_state_with_message')

    async def get_state(self) -> int:
        """
        Get the current state of the credential object
        :return: credential state of the object. Possible states:
                 2 - Request Sent
                 3 - Request Received
                 4 - Accepted

        Example:
        credential = await Credential.create(source_id, offer)
        credential.update_state()
        state = credential.get_state()
        """
        return await self._get_state(Credential, 'vcx_credential_get_state')

    def release(self) -> None:
        """
        Used to release memory associated with this object, used by the c library.
        :return:
        """
        self._release(Credential, 'vcx_credential_release')

    async def send_request(self, connection: Connection, payment_handle: int):
        """
        Approves the credential offer and submits a credential request. The result will be a credential stored in the prover's wallet.
        :param connection: connection to submit request from
        :param payment_handle: currently unused
        :return:
        Example:
        connection = await Connection.create(source_id)
        await connection.connect(phone_number)
        credential = await Credential.create(source_id, offer)
        await credential.send_request(connection, 0)
        """
        if not hasattr(Credential.send_request, "cb"):
            self.logger.debug("vcx_credential_send_request: Creating callback")
            Credential.send_request.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32))

        c_credential_handle = c_uint32(self.handle)
        c_connection_handle = c_uint32(connection.handle)
        c_payment = c_uint32(payment_handle)

        await do_call('vcx_credential_send_request',
                      c_credential_handle,
                      c_connection_handle,
                      c_payment,
                      Credential.send_request.cb)

    async def get_request_msg(self, my_pw_did: str, their_pw_did: str, payment_handle: int):
        """
        Approves the credential offer and gets the credential request message
        :param my_pw_did: my pairwaise did
        :param their_pw_did: pairwaise did of other side
        :param payment_handle: currently unused
        :return:
        Example:
        connection = await Connection.create(source_id)
        await connection.connect(phone_number)
        credential = await Credential.create(source_id, offer)
        await credential.get_request_msg(my_pw_did, their_pw_did, 0)
        """
        if not hasattr(Credential.get_request_msg, "cb"):
            self.logger.debug("vcx_credential_get_request_msg: Creating callback")
            Credential.get_request_msg.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)
        c_my_pw_did = c_char_p(my_pw_did.encode('utf-8'))
        c_their_pw_did = c_char_p(their_pw_did.encode('utf-8'))
        c_payment = c_uint32(payment_handle)

        msg = await do_call('vcx_credential_get_request_msg',
                      c_credential_handle,
                      c_my_pw_did,
                      c_their_pw_did,
                      c_payment,
                      Credential.get_request_msg.cb)

        return json.loads(msg.decode())

    async def get_payment_info(self):
        """
        Retrieve Payment Transaction Information for this Credential. Typically this will include
        how much payment is requried by the issuer, which needs to be provided by the prover, before the issuer will
        issue the credential to the prover. Ideally a prover would want to know how much payment is being asked before
        submitting the credential request (which triggers the payment to be made).
        Example:
        info = credential.get_payment_info()
        :return: payment information
             {
                 "payment_required":"one-time",
                 "payment_addr":"pov:null:OsdjtGKavZDBuG2xFw2QunVwwGs5IB3j",
                 "price":1
             }
        """
        if not hasattr(Credential.get_payment_info, "cb"):
            self.logger.debug("vcx_credential_get_payment_info: Creating callback")
            Credential.get_payment_info.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)
        data = await do_call('vcx_credential_get_payment_info',
                      c_credential_handle,
                      Credential.get_payment_info.cb)
        return json.loads(data.decode())

    async def get_payment_txn(self):
        """
        Retrieve the payment transaction associated with this credential. This can be used to get the txn that
        was used to pay the issuer from the prover.  This could be considered a receipt of payment from the payer to
        the issuer.
        :return: payment transaction
        {
            "amount":25,
            "inputs":[
                "pay:null:1_3FvPC7dzFbQKzfG"
            ],
            "outputs":[
                {"recipient":"pay:null:FrSVC3IrirScyRh","amount":5,"extra":null}
            ]
        }
        Example:
        txn = credential.get_payment_txn()
        """
        if not hasattr(Credential.get_payment_txn, "cb"):
            self.logger.debug("vcx_credential_get_payment_txn: Creating callback")
            Credential.get_payment_txn.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_credential_handle = c_uint32(self.handle)

        payment_txn = await do_call('vcx_credential_get_payment_txn',
                      c_credential_handle,
                      Credential.get_payment_txn.cb)

        return json.loads(payment_txn.decode())











import json

LIBRARY = "libvcx.so"
from vcx.cdll import _cdll
from enum import IntEnum
from ctypes import c_char_p, byref, c_uint32
import logging
from typing import Optional


class ErrorCode(IntEnum):
    Success = 0,
    IndyInvalidWalletHandle = 200,
    IndyWalletRecordNotFound = 212,
    IndyDuplicateWalletRecord = 213,
    UnknownError = 1001,
    ConnectionError = 1002,
    InvalidConnectionHandle = 1003,
    InvalidConfiguration = 1004,
    NotReady = 1005,
    NoEndpoint = 1006,
    InvalidOption = 1007,
    InvalidDid = 1008,
    InvalidVerkey = 1009,
    CouldNotConnect = 1010,
    InvalidNonce = 1011,
    InvalidKeyDelegate = 1012,
    InvalidUrl = 1013,
    NotBase58 = 1014,
    InvalidIssuerCredentialHandle = 1015,
    InvalidJson = 1016,
    InvalidProofHandle = 1017,
    InvalidCredentialRequest = 1018,
    InvalidMessagePack = 1019,
    InvalidMessages = 1020,
    InvalidAttributesStructure = 1021,
    BigNumberError = 1022,
    InvalidProof = 1023,
    InvalidGenesisTxnPath = 1024,
    PoolLedgerConnect = 1025,
    CreatePoolConifg = 1026,
    InvalidProofCredentialData = 1027,
    IndySubmitRequestError = 1028,
    BuildCredentialDefReqErr = 1029,
    NoPoolOpen = 1030,
    InvalidSchema = 1031,
    FailedPoolCompliance = 1032,
    InvalidHttpResponse = 1033,
    CreateCredentialDefErr = 1034,
    UnknownLibindyError = 1035,
    InvalidCredentialDef = 1036,
    InvalidCredentialDefHandle = 1037,
    TimeoutLibindyError = 1038,
    CredentialDefAlreadyCreated = 1039,
    InvalidSchemaSequenceNumber = 1040,
    InvalidSchemaCreation = 1041,
    InvalidSchemaHandle = 1042,
    InvalidMasterSecret = 1043,
    AlreadyInitialized = 1044,
    InvalidInviteDetails = 1045,
    InvalidSelfAttestedVal = 1046,
    InvalidPredicate = 1047,
    InvalidObjHandle = 1048,
    InvalidDisclosedProofHandle = 1049,
    SerializationError = 1050,
    WalletAlreadyExists = 1051,
    WalletAlreadyOpen = 1052,
    InvalidCredentialHandle = 1053,
    InvalidCredentialJson = 1054,
    CreateCredentialFailed = 1055,
    CreateProofError = 1056,
    InvalidWalletHandle = 1057,
    InvalidWalletCreation = 1058,
    InvalidPoolName = 1059,
    CannotDeleteConnection = 1060,
    CreateConnectionError = 1061,
    InvalidWalletSetup = 1062,
    CommonError = 1063,
    InsufficientTokenAmount = 1064,
    UnknownTxnType = 1065,
    InvalidPaymentAddress = 1066,
    InvalidLibindyParam = 1067,
    InvalidPayment = 1068,
    MissingWalletKey = 1069,
    ObjectCacheError = 1070,
    NoPaymentInformation = 1071,
    DuplicateWalletRecord = 1072,
    WalletRecordNotFound = 1073,
    IOError = 1074,
    WalletAccessFailed = 1075,
    MissingWalletName = 1076,
    MissingExportedWalletPath = 1077,
    MissingBackupKey = 1078,
    WalletNotFound = 1079,
    LibindyInvalidStructure = 1080,
    InvalidState = 1081,
    InvalidLedgerResponse = 1082,
    DidAlreadyExistsInWallet = 1083,
    DuplicateMasterSecret = 1084,
    ThreadError = 1085,
    InvalidProofRequest = 1086,
    MissingPaymentMethod = 1087,
    DuplicateSchema = 1088,
    UnknownLibindyRejection = 1089,
    LoggingError = 1090
    InvalidRevocationDetails = 1091,
    InvalidRevEntry = 1092,
    InvalidRevocationTimestamp = 1093,
    UnknownSchemaRejection = 1094,
    InvalidRevRegDefCreation = 1095,
    # EC 1096 - 1099 are reserved for proprietary forks of libVCX
    InvalidAttachmentEncoding = 1100,
    UnknownAttachmentEncoding = 1101,
    UnknownMimeType = 1102,
    ActionNotSupported = 1103,
    InvalidRedirectDetails = 1104,
    NoAgentInfo = 1106


class VcxError(Exception):
    # error_code: ErrorCode
    # error_msg: Optional[str] - human-readable error description
    # sdk_error_full_message: Optional[str] - vcx full error message.
    # sdk_error_cause: Optional[str] - vcx error cause.
    # sdk_error_backtrace: Optional[str] - vcx error backtrace.
    #   Collecting of backtrace can be enabled by setting environment variable `RUST_BACKTRACE=1`

    def __init__(self, error_code: ErrorCode, error_details: Optional[dict] = None):
        self.error_code = error_code
        self.error_msg = error_message(error_code)
        if error_details:
            self.error_msg = error_details['error']
            self.sdk_error_full_message = error_details['message']
            self.sdk_error_cause = error_details['cause']
            self.sdk_error_backtrace = error_details['backtrace']


def error_message(error_code: int) -> str:
    logger = logging.getLogger(__name__)
    name = 'vcx_error_c_message'
    c_error_code = c_uint32(error_code)
    getattr(_cdll(), name).restype = c_char_p
    err_msg = getattr(_cdll(), name)(c_error_code)
    logger.debug("error_message: Function %s[%s] returned error_message: %s", name, error_code, err_msg)
    return err_msg.decode()


def get_error_details() -> Optional[dict]:
    logger = logging.getLogger(__name__)
    logger.debug("get_error_details: >>>")

    error_c = c_char_p()
    getattr(_cdll(), 'vcx_get_current_error')(byref(error_c))
    error_details = json.loads(error_c.value.decode()) if error_c.value else None

    logger.debug("get_error_details: <<< error_details: %s", error_details)
    return error_details



from vcx.common import do_call, create_cb, release
from ctypes import *

import logging
import json



class VcxBase:
    HANDLES = {}

    def __init__(self, source_id: str):
        self._source_id = source_id
        self._logger = logging.getLogger(__name__)
        self._handle = 0

    @property
    def handle(self):
        return self._handle

    @handle.setter
    def handle(self, handle):
        self._handle = handle

    @property
    def source_id(self):
        return self._source_id

    @source_id.setter
    def source_id(self, x):
        self._source_id = x

    @property
    def logger(self):
        return self._logger

    @classmethod
    async def _create(cls, fn: str, constructor_args, c_args):
        obj = cls(*constructor_args)

        if not hasattr(cls.create, "cb"):
            obj.logger.debug("{}: Creating callback".format(fn))
            cls.create.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))

        obj.handle = await do_call(fn,
                                   *c_args,
                                   cls.create.cb)

        VcxBase.HANDLES[obj.handle] = obj

        obj.logger.debug("created {} object".format(cls))

        return obj

    @classmethod
    async def _deserialize(cls, fn: str, data: str, *args):
        obj = cls(*args)

        if not hasattr(cls.deserialize, "cb"):
            obj.logger.debug("{}: Creating callback".format(fn))
            cls.deserialize.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_uint32))

        c_data = c_char_p(data.encode('utf-8'))

        obj.handle = await do_call(fn,
                                   c_data,
                                   cls.deserialize.cb)

        if obj.handle not in VcxBase.HANDLES:
            obj.logger.debug("deserialized {} object".format(cls))
            VcxBase.HANDLES[obj.handle] = obj
            return obj
        else:
            obj.logger.debug("duplicate {} object deserialized".format(cls))
            obj_to_return = VcxBase.HANDLES[obj.handle]
            obj.handle = 0
            return obj_to_return

    async def _serialize(self, cls, fn_str: str) -> dict:
        if not hasattr(cls.serialize, "cb"):
            self.logger.debug("{}: Creating callback".format(fn_str))
            cls.serialize.cb = create_cb(CFUNCTYPE(None, c_uint32, c_uint32, c_char_p))

        c_handle = c_uint32(self.handle)

        data = await do_call(fn_str,
                             c_handle,
                             cls.serialize.cb)

        self.logger.debug("serialized {} object".format(cls))
        j = json.loads(data.decode())
        return j

    def _release(self, cls, fn: str):
        self.logger.debug("Releasing %s handle: %s", cls, self.handle)
        if self.handle in VcxBase.HANDLES:
            c_handle = c_uint32(self.handle)
            VcxBase.HANDLES.pop(self.handle)
            release(fn, c_handle)
